import sys
import locale

reload(sys)

def main():
  encoding = locale.getdefaultlocale()[1]
  if not encoding:
    return False

  sys.setdefaultencoding(encoding)
  textmap = {
    'cp936': u'\u4e2d\u6587',
    'cp1252': u'Lat\u012Bna',
    'cp932': u'\u306b\u307b\u3093\u3054'
  }
  if textmap.has_key(encoding):
    print textmap[encoding]
  return True

if __name__ == '__main__':
  print main()



#  Copyright (c) 2019 JD Williams
#
#  This file is part of Firefly, a Python SOA framework built by JD Williams. Firefly is free software; you can
#  redistribute it and/or modify it under the terms of the GNU General Public License as published by the
#  Free Software Foundation; either version 3 of the License, or (at your option) any later version.
#
#  Firefly is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
#  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
#  Public License for more details. You should have received a copy of the GNU Lesser General Public
#  License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#  You should have received a copy of the GNU General Public License along with Firefly. If not, see
#  <http://www.gnu.org/licenses/>.

from __future__ import annotations

from typing import List, Optional, Union

import firefly.domain as ffd

from ..messaging.system_bus import SystemBusAware
from ...entity.core.user import User

CATEGORIES = ('read', 'write', 'admin')


class Kernel(SystemBusAware):
    user: Optional[User] = None
    required_scopes: Optional[List[str]] = None
    http_request: Optional[dict] = None
    secured: Optional[bool] = None

    def reset(self):
        self.user = None
        self.required_scopes = None
        self.http_request = None
        self.secured = None

    @property
    def is_authorized(self):
        if self.required_scopes is None or len(self.required_scopes) == 0:
            return True  # No required scopes, return True

        if self.user is None:
            return self.secured is False

        if len(self.user.scopes) > 0:
            for scope in self.required_scopes:
                for user_scope in self.user.scopes:
                    if self._has_grant(scope, user_scope):
                        return True

        return False

    @staticmethod
    def _has_grant(scope: str, user_scope: str):
        parts = scope.lower().split('.')
        user = user_scope.lower().split('.')

        for i, part in enumerate(parts):
            if i >= len(user):
                return False

            if user[i] == 'admin':
                return True

            if part not in CATEGORIES and part != user[i]:
                return False

            if part in CATEGORIES:
                if user[i] not in CATEGORIES:
                    return False
                if part == 'admin':
                    return False
                if part == 'write':
                    return user[i] == 'write'
                if part == 'read':
                    return user[i] in ('read', 'write')

        return True

    def has_tenant(self):
        return self.user is not None and self.user.tenant is not None

    def reject_missing_tenant(self):
        if not self.has_tenant():
            raise ffd.Unauthorized()

    def reject_if_missing(self, scopes: Union[str, List[str]]):
        if self.user is None:
            raise ffd.Unauthorized()

        if isinstance(scopes, str):
            scopes = [scopes]
        for my_scope in self.user.scopes:
            reject = True
            for scope in scopes:
                if self._has_grant(my_scope, scope):
                    reject = False
                    break
            if reject:
                raise ffd.Unauthorized()

    def boot(self):
        self.invoke(ffd.LoadContainers())



#  Copyright (c) 2019 JD Williams
#
#  This file is part of Firefly, a Python SOA framework built by JD Williams. Firefly is free software; you can
#  redistribute it and/or modify it under the terms of the GNU General Public License as published by the
#  Free Software Foundation; either version 3 of the License, or (at your option) any later version.
#
#  Firefly is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
#  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
#  Public License for more details. You should have received a copy of the GNU Lesser General Public
#  License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#  You should have received a copy of the GNU General Public License along with Firefly. If not, see
#  <http://www.gnu.org/licenses/>.

from __future__ import annotations

import os
from typing import Callable

import firefly.domain as ffd
from firefly.domain.service.logging.logger import LoggerAware
from firefly.domain.service.messaging.middleware import Middleware


class LoggingMiddleware(Middleware, LoggerAware):
    def __init__(self, message: str = None):
        self._message = message or 'Message added to bus: %s'

    def __call__(self, message: ffd.Message, next_: Callable) -> ffd.Message:
        original_log_level = None
        if ('DEBUG' in os.environ and os.environ['DEBUG']) or ('debug' in message.headers and message.headers['debug']):
            original_log_level = self._logger.get_level()
            self._logger.set_level_to_debug()

        self.info(self._message, message)
        try:
            self.debug('Message properties: %s', message.to_dict())
        except AttributeError:
            pass
        ret = next_(message)
        self.debug('Response: %s', str(ret))

        if original_log_level is not None:
            self._logger.set_level(original_log_level)

        return ret



#  Copyright (c) 2019 JD Williams
#
#  This file is part of Firefly, a Python SOA framework built by JD Williams. Firefly is free software; you can
#  redistribute it and/or modify it under the terms of the GNU General Public License as published by the
#  Free Software Foundation; either version 3 of the License, or (at your option) any later version.
#
#  Firefly is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
#  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
#  Public License for more details. You should have received a copy of the GNU Lesser General Public
#  License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#  You should have received a copy of the GNU General Public License along with Firefly. If not, see
#  <http://www.gnu.org/licenses/>.

from __future__ import annotations

import inspect
from abc import ABC, abstractmethod
from dataclasses import fields
from typing import Type, get_type_hints, List, Union, Callable, Dict, Tuple

import firefly.domain as ffd
import inflection
# noinspection PyDataclass
from jinjasql import JinjaSql

from .abstract_storage_interface import AbstractStorageInterface
from .rdb_repository import Index, Column


# noinspection PyDataclass
class RdbStorageInterface(AbstractStorageInterface, ABC):
    _serializer: ffd.Serializer = None
    _registry: ffd.Registry = None
    _j: JinjaSql = None
    _cache: dict = {}
    _sql_prefix = 'sql'
    _map_indexes = False
    _map_all = False
    _identifier_quote_char = '"'

    def __init__(self, **kwargs):
        self._tables_checked = []

    def _add(self, entity: Union[ffd.Entity, List[ffd.Entity]]):
        entities = entity
        if not isinstance(entity, list):
            entities = [entity]

        return self._execute(*self._generate_query(
            entity,
            f'{self._sql_prefix}/insert.sql',
            {'data': list(map(self._data_fields, entities))}
        ))

    def _generate_select(self, entity_type: Type[ffd.Entity], criteria: ffd.BinaryOp = None, limit: int = None,
                         offset: int = None, sort: Tuple[Union[str, Tuple[str, bool]]] = None, count: bool = False):
        indexes = [f.name for f in fields(entity_type)
                   if f.metadata.get('index') is True or f.metadata.get('id') is True]
        data = {
                'columns': self._select_list(entity_type),
                'count': count,
            }
        if criteria is not None:
            data['criteria'] = criteria

        if sort is not None:
            sort_fields = []
            for s in sort:
                if str(s[0]) in indexes or self._map_indexes is False:
                    sort_fields.append(s)
            data['sort'] = sort_fields

        if limit is not None:
            data['limit'] = limit

        if offset is not None:
            data['offset'] = offset

        return self._generate_query(entity_type, f'{self._sql_prefix}/select.sql', data)

    def _all(self, entity_type: Type[ffd.Entity], criteria: ffd.BinaryOp = None, limit: int = None, offset: int = None,
             sort: Tuple[Union[str, Tuple[str, bool]]] = None, raw: bool = False, count: bool = False):
        sql, params = self._generate_select(
            entity_type, criteria, limit=limit, offset=offset, sort=sort, count=count
        )
        results = self._execute(sql, params)

        ret = []
        if count:
            return results[0]['c']

        for row in results:
            self.debug('Result row: %s', dict(row))
            ret.append(self._build_entity(entity_type, row, raw=raw))

        return ret

    def _find(self, uuid: str, entity_type: Type[ffd.Entity]):
        results = self._execute(*self._generate_query(
            entity_type,
            f'{self._sql_prefix}/select.sql',
            {
                'columns': self._select_list(entity_type),
                'criteria': ffd.Attr(entity_type.id_name()) == uuid
            }
        ))

        if len(results) == 0:
            return None

        if len(results) > 1:
            raise ffd.MultipleResultsFound()

        return self._build_entity(entity_type, results[0])

    def _remove(self, entity: Union[ffd.Entity, List[ffd.Entity], Callable]):
        criteria = entity
        if isinstance(entity, list):
            criteria = ffd.Attr(entity[0].id_name()).is_in([e.id_value() for e in entity])
        elif not isinstance(entity, ffd.BinaryOp):
            criteria = ffd.Attr(entity.id_name()) == entity.id_value()

        self._execute(*self._generate_query(entity, f'{self._sql_prefix}/delete.sql', {
            'criteria': criteria
        }))

    def _update(self, entity: ffd.Entity):
        criteria = ffd.Attr(entity.id_name()) == entity.id_value()
        try:
            criteria &= ffd.Attr('version') == getattr(entity, '__ff_version')
        except AttributeError:
            pass

        return self._execute(*self._generate_query(
            entity,
            f'{self._sql_prefix}/update.sql',
            {
                'data': self._data_fields(entity),
                'criteria': criteria
            }
        ))

    @abstractmethod
    def _ensure_connected(self):
        pass

    def clear(self, entity: Type[ffd.Entity]):
        self.execute(*self._generate_query(entity, f'{self._sql_prefix}/truncate_table.sql'))

    def destroy(self, entity: Type[ffd.Entity]):
        self.execute(*self._generate_query(entity, f'{self._sql_prefix}/drop_table.sql'))

    @staticmethod
    def _fqtn(entity: Type[ffd.Entity]):
        return inflection.tableize(entity.get_fqn())

    def _check_prerequisites(self, entity: Type[ffd.Entity]):
        self._ensure_connected()

    def get_entity_columns(self, entity: Type[ffd.Entity]):
        ret = []
        annotations_ = get_type_hints(entity)
        for f in fields(entity):
            if f.name.startswith('_'):
                continue

            c = Column(
                name=f.name,
                type=annotations_[f.name],
                length=f.metadata.get('length', None),
                is_id=f.metadata.get('id', False),
                is_indexed=f.metadata.get('index', False)
            )
            if self._map_all is True:
                ret.append(c)
            elif self._map_indexes and f.metadata.get('index', False):
                ret.append(c)
            elif f.metadata.get('id', False):
                ret.append(c)

        if not self._map_all:
            ret.insert(1, Column(name='document', type=dict))
            ret.insert(2, Column(name='__document', type=dict))
            ret.insert(3, Column(name='version', type=int, default=1))

        return ret

    def get_table_columns(self, entity: Type[ffd.Entity]):
        return self._get_table_columns(entity)

    @abstractmethod
    def _get_table_columns(self, entity: Type[ffd.Entity]):
        pass

    def add_column(self, entity: Type[ffd.Entity], column: Column):
        self.execute(
            *self._generate_query(entity, f'{self._sql_prefix}/add_column.sql', {'column': column})
        )

    def drop_column(self, entity: Type[ffd.Entity], column: Column):
        self.execute(
            *self._generate_query(entity, f'{self._sql_prefix}/drop_column.sql', {'column': column})
        )

    def get_entity_indexes(self, entity: Type[ffd.Entity]):
        ret = []
        table = self._fqtn(entity).replace('.', '_')
        for field_ in fields(entity):
            if 'index' in field_.metadata:
                if field_.metadata['index'] is True:
                    ret.append(
                        Index(table=table, columns=[field_.name], unique=field_.metadata.get('unique', False) is True)
                    )
                elif isinstance(field_.metadata['index'], str):
                    name = field_.metadata['index']
                    idx = next(filter(lambda i: i.name == name, ret))
                    if not idx:
                        ret.append(Index(table=table, columns=[field_.name],
                                         unique=field_.metadata.get('unique', False) is True))
                    else:
                        idx.columns.append(field_.name)
                        if field_.metadata.get('unique', False) is True and idx.unique is False:
                            idx.unique = True

        return ret

    def get_table_indexes(self, entity: Type[ffd.Entity]):
        return self._get_table_indexes(entity)

    @abstractmethod
    def _get_table_indexes(self, entity: Type[ffd.Entity]):
        pass

    def create_index(self, entity: Type[ffd.Entity], index: Index):
        self.execute(
            *self._generate_query(entity, f'{self._sql_prefix}/add_index.sql', {'index': index})
        )

    def drop_index(self, entity: Type[ffd.Entity], index: Index):
        self.execute(
            *self._generate_query(entity, f'{self._sql_prefix}/drop_index.sql', {'index': index})
        )

    def _build_entity(self, entity: Type[ffd.Entity], data, raw: bool = False):
        if raw is True:
            if self._map_all is True:
                return data
            else:
                return self._serializer.deserialize(data['document'])

        version = None
        if self._map_all is False:
            version = data['version']
            data = self._serializer.deserialize(data['document'])

        for k, v in self._get_relationships(entity).items():
            if v['this_side'] == 'one':
                data[k] = self._registry(v['target']).find(data[k])
            elif v['this_side'] == 'many':
                data[k] = list(self._registry(v['target']).filter(
                    lambda ee: getattr(ee, v['target'].id_name()).is_in(data[k])
                ))

        ret = entity.from_dict(data)
        if version is not None:
            setattr(ret, '__ff_version', version)
        return ret

    @staticmethod
    def _generate_index(name: str):
        return ''

    def execute(self, sql: str, params: dict = None):
        self._execute(sql, params)

    @abstractmethod
    def _execute(self, sql: str, params: dict = None):
        pass

    def _generate_query(self, entity: Union[ffd.Entity, List[ffd.Entity], Type[ffd.Entity]], template: str, params: dict = None):
        params = params or {}
        if isinstance(entity, list):
            entity = entity[0]
        if not inspect.isclass(entity):
            entity = entity.__class__

        def mapped_fields(e):
            return self.get_entity_columns(e)

        template = self._j.env.select_template([template, '/'.join(['sql', template.split('/')[1]])])
        data = {
            'fqtn': self._fqtn(entity),
            '_q': self._identifier_quote_char,
            'map_indexes': self._map_indexes,
            'map_all': self._map_all,
            'mapped_fields': mapped_fields,
            'indexes': list(map(lambda e: e.name, self.get_entity_indexes(entity))),
            'ids': entity.id_name() if isinstance(entity.id_name(), list) else [entity.id_name()],
            'field_types': {f.name: f.type for f in fields(entity)},
        }
        data.update(params)
        sql, params = self._j.prepare_query(template, data)

        return " ".join(sql.split()), params

    def create_table(self, entity_type: Type[ffd.Entity]):
        self.execute(
            *self._generate_query(
                entity_type,
                f'{self._sql_prefix}/create_table.sql',
                {'entity': entity_type, }
            )
        )

    def create_schema(self, entity_type: Type[ffd.Entity]):
        self.execute(
            *self._generate_query(
                entity_type,
                f'{self._sql_prefix}/create_schema.sql',
                {'context_name': entity_type.get_class_context()}
            )
        )

    def _data_fields(self, entity: ffd.Entity):
        ret = {}
        for f in self.get_entity_columns(entity.__class__):
            if f.name == 'document' and not hasattr(entity, 'document'):
                ret[f.name] = self._serialize_entity(entity)
            elif f.name == 'version':
                try:
                    ret['version'] = getattr(entity, '__ff_version')
                except AttributeError:
                    ret['version'] = 1
            elif inspect.isclass(f.type) and issubclass(f.type, ffd.AggregateRoot):
                ret[f.name] = getattr(entity, f.name).id_value()
            elif ffd.is_type_hint(f.type):
                origin = ffd.get_origin(f.type)
                args = ffd.get_args(f.type)
                if origin is List:
                    if issubclass(args[0], ffd.AggregateRoot):
                        ret[f.name] = self._serializer.serialize(
                            list(map(lambda e: e.id_value(), getattr(entity, f.name)))
                        )
                    else:
                        ret[f.name] = self._serializer.serialize(getattr(entity, f.name))
                elif origin is Dict:
                    if issubclass(args[1], ffd.AggregateRoot):
                        ret[f.name] = {k: v.id_value() for k, v in getattr(entity, f.name).items()}
                    else:
                        ret[f.name] = self._serializer.serialize(getattr(entity, f.name))
            elif f.type is list or f.type is dict:
                if hasattr(entity, f.name):
                    ret[f.name] = self._serializer.serialize(getattr(entity, f.name))
            else:
                ret[f.name] = getattr(entity, f.name)
                if isinstance(ret[f.name], ffd.ValueObject):
                    ret[f.name] = self._serializer.serialize(ret[f.name])
        return ret

    def _select_list(self, entity: Type[ffd.Entity]):
        if self._map_all:
            return list(map(lambda c: c.name, self.get_entity_columns(entity)))
        return ['document', 'version']



#  Copyright (c) 2019 JD Williams
#
#  This file is part of Firefly, a Python SOA framework built by JD Williams. Firefly is free software; you can
#  redistribute it and/or modify it under the terms of the GNU General Public License as published by the
#  Free Software Foundation; either version 3 of the License, or (at your option) any later version.
#
#  Firefly is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
#  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
#  Public License for more details. You should have received a copy of the GNU Lesser General Public
#  License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#  You should have received a copy of the GNU General Public License along with Firefly. If not, see
#  <http://www.gnu.org/licenses/>.

from typing import Callable

from firefly.presentation.web.polyfills import *  # __:skip
from firefly.presentation.web.js_libs.mithril import m

from firefly import SystemBus, Middleware, EventBus, domain as ffd, CommandBus, QueryBus, MessageFactory


# __pragma__('kwargs')
class LoggingMiddleware(Middleware):
    def __init__(self, prefix: str = ''):
        self.prefix = prefix

    def __call__(self, message: ffd.Message, next_: Callable) -> ffd.Message:
        console.log(self.prefix, message)
        return next_(message)


class CommandHandlingMiddleware(Middleware):
    def __call__(self, message: ffd.Message, next_: Callable) -> ffd.Message:
        return m.request({
            'method': 'POST',
            'url': f'{process.env.FF_HOST}/{message.get_context()}',
            'body': message.to_dict(),
        })


class QueryHandlingMiddleware(Middleware):
    def __call__(self, message: ffd.Message, next_: Callable) -> ffd.Message:
        return m.request({
            'method': 'GET',
            'url': f'{process.env.FF_HOST}/{message.get_context()}',
            'params': {
                'query': JSON.stringify(message.to_dict()),
            },
        })
# __pragma__('nokwargs')


message_factory = MessageFactory()
bus = SystemBus()
bus._event_bus = EventBus([
    LoggingMiddleware('Event Bus'),
])
# noinspection PyProtectedMember
bus._event_bus._message_factory = message_factory

bus._command_bus = CommandBus([
    LoggingMiddleware('Command Bus'),
    CommandHandlingMiddleware(),
])
# noinspection PyProtectedMember
bus._command_bus._message_factory = message_factory

bus._query_bus = QueryBus([
    LoggingMiddleware('Query Bus'),
    QueryHandlingMiddleware(),
])
# noinspection PyProtectedMember
bus._query_bus._message_factory = message_factory



#  Copyright (c) 2019 JD Williams
#
#  This file is part of Firefly, a Python SOA framework built by JD Williams. Firefly is free software; you can
#  redistribute it and/or modify it under the terms of the GNU General Public License as published by the
#  Free Software Foundation; either version 3 of the License, or (at your option) any later version.
#
#  Firefly is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
#  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
#  Public License for more details. You should have received a copy of the GNU Lesser General Public
#  License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#  You should have received a copy of the GNU General Public License along with Firefly. If not, see
#  <http://www.gnu.org/licenses/>.

from __future__ import annotations

import importlib
import inspect
from typing import List

import firefly.domain as ffd


class LoadPresentationLayer(ffd.ApplicationService):
    _context_map: ffd.ContextMap = None

    def __call__(self, **kwargs):
        for context in self._context_map.contexts:
            for obj in self._load_module(context):
                context.ports.append(obj)

        self.dispatch(ffd.PresentationLayerLoaded())

    @staticmethod
    def _load_module(context: ffd.Context) -> List[object]:
        module_name = context.config.get('ui_module', '{}.ui')
        try:
            module = importlib.import_module(module_name.format(context.name))
        except (ModuleNotFoundError, KeyError):
            return []

        ret = []
        for k, v in module.__dict__.items():
            if not inspect.isclass(v):
                continue
            if hasattr(v, '__ff_port'):
                ret.append(v)

        return ret



#  Copyright (c) 2019 JD Williams
#
#  This file is part of Firefly, a Python SOA framework built by JD Williams. Firefly is free software; you can
#  redistribute it and/or modify it under the terms of the GNU General Public License as published by the
#  Free Software Foundation; either version 3 of the License, or (at your option) any later version.
#
#  Firefly is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
#  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
#  Public License for more details. You should have received a copy of the GNU Lesser General Public
#  License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#  You should have received a copy of the GNU General Public License along with Firefly. If not, see
#  <http://www.gnu.org/licenses/>.

# __pragma__('skip')
from __future__ import annotations

import inspect
import typing
from dataclasses import fields
from datetime import datetime, date
from typing import List, Union, Dict

import inflection
from firefly.domain.entity.validation import IsValidEmail, HasLength, MatchesPattern, IsValidUrl, IsLessThanOrEqualTo, \
    IsLessThan, IsGreaterThanOrEqualTo, IsGreaterThan, IsMultipleOf, HasMaxLength, HasMinLength, parse
from firefly.domain.meta.build_argument_list import build_argument_list
from firefly.domain.meta.entity_meta import EntityMeta
from firefly.domain.utils import is_type_hint, get_origin, get_args, can_be_type

from .event_buffer import EventBuffer
from .generic_base import GenericBase
from .parameter import Parameter


class Empty:
    pass


_defs = {}


# noinspection PyDataclass
class ValueObject(metaclass=EntityMeta):
    _logger = None
    _mappings = {
        str: 'string',
        int: 'integer',
        float: 'number',
        bool: 'boolean',
        datetime: 'string',
        date: 'string',
    }

    def __init__(self, **kwargs):
        pass

    def to_dict(self, skip: list = None, force_all: bool = False):
        ret = {}
        annotations_ = typing.get_type_hints(self.__class__)
        for field_ in fields(self):
            if field_.name.startswith('_'):
                continue
            if field_.metadata.get('internal') is True and force_all is False:
                continue

            type_ = annotations_[field_.name]
            if inspect.isclass(type_) and issubclass(type_, ValueObject):
                f = getattr(self, field_.name)
                if isinstance(f, ValueObject):
                    ret[field_.name] = f.to_dict()
                else:
                    ret[field_.name] = None
            elif is_type_hint(annotations_[field_.name]):
                origin = get_origin(type_)
                args = get_args(type_)
                if origin is List and can_be_type(args[0], ValueObject):
                    if getattr(self, field_.name) is None:
                        ret[field_.name] = None
                    else:
                        ret[field_.name] = list(map(
                            lambda v: v.to_dict() if isinstance(v, ValueObject) else None,
                            getattr(self, field_.name)
                        ))
                elif origin is Dict and can_be_type(args[1], ValueObject):
                    if getattr(self, field_.name) is None:
                        ret[field_.name] = None
                    else:
                        ret[field_.name] = {k: v.to_dict() for k, v in getattr(self, field_.name).items()}
                else:
                    ret[field_.name] = getattr(self, field_.name)
            else:
                ret[field_.name] = getattr(self, field_.name)

        if skip is not None:
            d = ret.copy()
            for k in ret.keys():
                if k in skip:
                    del d[k]
            return d

        return ret

    def load_dict(self, d: dict):
        data = build_argument_list(d.copy(), self.__class__, strict=False)
        t = typing.get_type_hints(self.__class__)
        for name, type_ in t.items():
            if name in data:
                if inspect.isclass(type_) and issubclass(type_, (datetime, date)) and isinstance(data[name], str):
                    setattr(self, name, parse(data[name], ignoretz=True))
                elif inspect.isclass(type_) and issubclass(type_, ValueObject) and isinstance(data[name], type_):
                    existing = getattr(self, name)
                    if existing is None:
                        setattr(self, name, data[name])
                    elif name in d and isinstance(d[name], dict):
                        existing.load_dict(d[name])
                else:
                    try:
                        if data[name] is not None and not isinstance(data[name], type_):
                            setattr(self, name, type_(data[name]))
                        else:
                            setattr(self, name, data[name])
                    except TypeError:
                        setattr(self, name, data[name])

    @classmethod
    def get_dto_schema(cls, stack: List[type] = None):
        global _defs

        stack = stack or []
        if len(stack) == 0:
            _defs = {}

        if cls in stack or cls.__name__ in _defs:
            if cls.__name__ not in _defs:
                _defs[cls.__name__] = {}
            return _defs[cls.__name__]

        stack.append(cls)

        ret = {
            '$schema': 'http://json-schema.org/draft-07/schema#',
            'title': cls.__name__,
            'type': 'object',
        }

        types_ = typing.get_type_hints(cls)
        props = {}
        required_fields = []
        for field_ in fields(cls):
            if field_.name.startswith('_'):
                continue

            if 'hidden' in field_.metadata and field_.metadata['hidden'] is True:
                continue

            prop = {
                'title': field_.metadata.get('title') or inflection.humanize(field_.name),
            }
            t = types_[field_.name]
            if t in cls._mappings:
                if t in cls._mappings:
                    prop['type'] = cls._mappings[t]

            if is_type_hint(t):
                prop = cls._process_type_hint(t, prop, stack)
            elif inspect.isclass(t) and issubclass(t, ValueObject):
                prop['type'] = 'object'
                subclasses = t.__subclasses__()
                if len(subclasses):
                    prop['items'] = {'oneOf': []}
                    for subclass in subclasses:
                        subclass.get_dto_schema(stack.copy())
                        prop['items']['oneOf'].append({'$ref': f'#/definitions/{subclass.__name__}'})
                else:
                    t.get_dto_schema(stack.copy())
                    prop = {'$ref': f'#/definitions/{t.__name__}'}

            if 'validators' in field_.metadata:
                for validator in field_.metadata['validators']:
                    if isinstance(validator, IsValidEmail):
                        prop['format'] = 'email'
                    elif isinstance(validator, HasLength):
                        prop['minLength'] = validator.length
                        prop['maxLength'] = validator.length
                    elif isinstance(validator, MatchesPattern):
                        prop['pattern'] = validator.regex
                    elif isinstance(validator, IsValidUrl):
                        prop['format'] = 'uri'
                    elif isinstance(validator, IsLessThanOrEqualTo):
                        prop['maximum'] = validator.value
                    elif isinstance(validator, IsLessThan):
                        prop['maximum'] = validator.value
                        prop['exclusiveMaximum'] = True
                    elif isinstance(validator, IsGreaterThanOrEqualTo):
                        prop['minimum'] = validator.value
                    elif isinstance(validator, IsGreaterThan):
                        prop['minimum'] = validator.value
                        prop['exclusiveMinimum'] = True
                    elif isinstance(validator, IsMultipleOf):
                        prop['multipleOf'] = validator.value
                    elif isinstance(validator, HasMaxLength):
                        prop['maxLength'] = validator.length
                    elif isinstance(validator, HasMinLength):
                        prop['minLength'] = validator.length

            if t is datetime:
                prop['format'] = 'date-time'
            elif t is date:
                prop['format'] = 'date'

            if 'format' in field_.metadata:
                prop['format'] = field_.metadata.get('format')

            if 'required' in field_.metadata and field_.metadata['required'] is True:
                try:
                    if isinstance(field_.default_factory(), Empty):
                        required_fields.append(field_.name)
                except TypeError:
                    required_fields.append(field_.name)

            else:
                prop['default'] = None

            props[field_.name] = prop

        ret['properties'] = props
        if len(required_fields) > 0:
            ret['required'] = required_fields

        if len(stack) == 1:
            ret['definitions'] = _defs
        else:
            _defs[cls.__name__] = ret
            return {'$ref': f'#/definitions/{cls.__name__}'}

        return ret

    @classmethod
    def _process_type_hint(cls, obj, config: dict, stack: list):
        if not is_type_hint(obj):
            return config

        if get_origin(obj) is List:
            config['type'] = 'array'
            args = get_args(obj)
            if inspect.isclass(args[0]) and issubclass(args[0], ValueObject):
                subclasses = args[0].__subclasses__()
                if len(subclasses):
                    config['items'] = {'oneOf': []}
                    for subclass in subclasses:
                        subclass.get_dto_schema(stack.copy())
                        config['items']['oneOf'].append({'$ref': f'#/definitions/{subclass.__name__}'})
                else:
                    args[0].get_dto_schema(stack.copy())
                    config['items'] = {'$ref': f'#/definitions/{args[0].__name__}'}

            elif is_type_hint(args[0]):
                config['items'] = cls._process_type_hint(args[0], {}, stack)['items']

            elif args[0] in cls._mappings:
                config['items'] = {
                    'type': cls._mappings[args[0]]
                }

        elif get_origin(obj) is Dict:
            args = get_args(obj)
            config['type'] = 'object'
            if 'properties' in config:
                del config['properties']

            if is_type_hint(args[1]):
                options = cls._process_type_hint(args[1], {}, stack)
            elif inspect.isclass(args[1]) and issubclass(args[1], ValueObject):
                args[1].get_dto_schema(stack.copy())
                options = {'$ref': f'#/definitions/{args[1].__name__}'}
            else:
                options = {'type': cls._mappings[args[1]]}

            config['patternProperties'] = {
                r'.*': options
            }

        elif get_origin(obj) is Union:
            args = get_args(obj)
            config['items'] = {'oneOf': []}
            for arg in args:
                if is_type_hint(arg):
                    raise NotImplementedError()
                elif inspect.isclass(arg) and issubclass(arg, ValueObject):
                    arg.get_dto_schema(stack.copy())
                    config['items']['oneOf'].append({'$ref': f'#/definitions/{arg.__name__}'})
                elif arg in cls._mappings:
                    config['items']['oneOf'].append({'type': cls._mappings[arg]})

        return config

    def debug(self, *args, **kwargs):
        return self._logger.debug(*args, **kwargs)

    def info(self, *args, **kwargs):
        return self._logger.info(*args, **kwargs)

    def warning(self, *args, **kwargs):
        return self._logger.warning(*args, **kwargs)

    def error(self, *args, **kwargs):
        return self._logger.error(*args, **kwargs)

    @classmethod
    def from_dict(cls, data: dict, map_: dict = None, skip: list = None):
        if map_ is not None:
            d = data.copy()
            for source, target in map_.items():
                if source in d:
                    d[target] = d[source]
            data = d

        if skip is not None:
            d = data.copy()
            for k in data.keys():
                if k in skip:
                    del d[k]
            data = d

        return cls(**build_argument_list(data, cls))
    # __pragma__('noskip')



#  Copyright (c) 2019 JD Williams
#
#  This file is part of Firefly, a Python SOA framework built by JD Williams. Firefly is free software; you can
#  redistribute it and/or modify it under the terms of the GNU General Public License as published by the
#  Free Software Foundation; either version 3 of the License, or (at your option) any later version.
#
#  Firefly is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
#  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
#  Public License for more details. You should have received a copy of the GNU Lesser General Public
#  License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#  You should have received a copy of the GNU General Public License along with Firefly. If not, see
#  <http://www.gnu.org/licenses/>.
#
#  This file is part of Firefly, a Python SOA framework built by JD Williams. Firefly is free software; you can
#  redistribute it and/or modify it under the terms of the GNU General Public License as published by the
#  Free Software Foundation; either version 3 of the License, or (at your option) any later version.
#
#  Firefly is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
#  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
#  Public License for more details. You should have received a copy of the GNU Lesser General Public
#  License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#  You should have received a copy of the GNU General Public License along with Firefly. If not, see
#  <http://www.gnu.org/licenses/>.

from datetime import datetime, date

# __pragma__('kwargs')
# __pragma__('opov')
# __pragma__('js', '{}', "const uuidv1 = require('uuid/v1');")


document = {}
window = {}
process = {}
JSON = {}
moment = {}
navigator = {}
this = {}  # __:skip


# __pragma__('skip')
def uuidv1():
    pass


def require(*args, **kwargs):
    pass
# __pragma__('noskip')


class Uuid:
    def __init__(self):
        self.uuid1 = uuidv1


uuid = Uuid()


class Console:
    def log(self, *args):
        pass

    def error(self, *args):
        pass


console = Console()  # __:skip


class ABC:
    pass


def abstractmethod(cls):
    return cls


class ABCMeta:
    pass


class MessageMeta:
    pass


def dataclass(_cls=None, *, init=True, repr=True, eq=True, order=False,
              unsafe_hash=False, frozen=False):
    return _cls
    if _cls is None:
        def wrapper(_cls):
            return _do_make_dataclass(_cls, init=init, repr=repr, eq=eq, order=order, unsafe_hash=unsafe_hash,
                                      frozen=frozen)

        return wrapper

    return _do_make_dataclass(_cls, init=init, repr=repr, eq=eq, order=order, unsafe_hash=unsafe_hash,
                              frozen=frozen)


def _do_make_dataclass(_cls, *, init=True, repr=True, eq=True, order=False,
              unsafe_hash=False, frozen=False):
    if init:
        def __init(self, **kwargs):
            for k, v in kwargs.items():
                if hasattr(self, k):
                    setattr(self, k, v)

        _cls['__init__'] = __init

    return _cls


def fields(*args, **kwargs):
    console.log('fields() called')


class MissingType:
    pass


MISSING = MissingType()


def field(*, default=MISSING, default_factory=MISSING, init=True, repr=True,
          hash=None, compare=True, metadata=None):
    return {
        'default': default,
        'default_factory': default_factory,
        'init': init,
        'repr': repr,
        'hash': hash,
        'compare': compare,
        'metadata': metadata,
    }


class Field:
    pass


def asdict():
    pass


def is_dataclass():
    pass


def make_dataclass(cls_name, fields, *, bases=(), namespace=None, init=True,
                   repr=True, eq=True, order=False, unsafe_hash=False,
                   frozen=False):
    return type.__new__(type, cls_name, bases, fields)


class DataClassBase:
    def __init__(self, **kwargs):
        for k, v in kwargs.items():
            if hasattr(self, k) and not callable(getattr(self, k)):
                # TODO Figure out why setattr doesn't work with "name." Could be a bug in transcrypt.
                if k == 'name':
                    self.name = v
                else:
                    setattr(self, k, v)

    def to_dict(self):
        ret = {}
        for k in dir(self):
            if not k.startswith('__') and not callable(getattr(self, k)):
                ret[k] = getattr(self, k)

        return ret


class Entity(DataClassBase):
    def __init__(self, **kwargs):
        for k in dir(self):
            if k.startswith('__') or not isinstance(getattr(self, k), str):
                continue
            if getattr(self, k) == '__gen_uuid__':
                kwargs[k] = uuidv1()
        super().__init__(**kwargs)


class AggregateRoot(Entity):
    pass


class ValueObject(DataClassBase):
    pass


class Message(DataClassBase):
    headers: dict = {}
    _id: str = None
    _context: str = None
    _name: str = None
    _type: str = None

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self._id = uuidv1()

    def get_context(self):
        return self._context

    def to_dict(self):
        ret = super().to_dict()
        if isinstance(self, Command):
            ret['_type'] = 'command'
        elif isinstance(self, Event):
            ret['_type'] = 'event'
        else:
            ret['_type'] = 'query'

        ret['_name'] = self.__class__.__name__

        return ret


class Command(Message):
    pass


class Query(Message):
    pass


class Event(Message):
    pass


def uuid1():
    return 'TODO: Implement JS uuid'


uuid = {
    'uuid1': uuid1
}


class TypeVar:
    def __init__(self, *args, **kwargs):
        pass


Union = {}
Type = {}
Tuple = {}
List = {}
Any = {}


def get_type_hints():
    pass


inflection = {}



#  Copyright (c) 2019 JD Williams
#
#  This file is part of Firefly, a Python SOA framework built by JD Williams. Firefly is free software; you can
#  redistribute it and/or modify it under the terms of the GNU General Public License as published by the
#  Free Software Foundation; either version 3 of the License, or (at your option) any later version.
#
#  Firefly is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
#  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
#  Public License for more details. You should have received a copy of the GNU Lesser General Public
#  License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#  You should have received a copy of the GNU General Public License along with Firefly. If not, see
#  <http://www.gnu.org/licenses/>.

from __future__ import annotations

import inspect
import typing
from datetime import datetime, date

import inflection

from .validation import *
from firefly.domain.value_object import ValueObject, Empty
from firefly.domain.meta.context_aware import ContextAware
from firefly.domain.meta.build_argument_list import build_argument_list
from firefly.domain.utils import is_type_hint, get_origin, get_args

# __pragma__('skip')
import uuid

from dataclasses import is_dataclass, fields, field, MISSING
from typing import List, Callable
from abc import ABC
# __pragma__('noskip')
# __pragma__ ('ecom')
"""?
from firefly.presentation.web.polyfills import is_dataclass, fields, field, MISSING, asdict, List, Callable, uuid
?"""
# __pragma__ ('noecom')
# __pragma__('kwargs')


# noinspection PyDataclass
class Entity(ContextAware, ValueObject):
    _logger = None

    def __init__(self, **kwargs):
        pass

    def __post_init__(self):
        if is_dataclass(self):
            missing = []
            for field_ in fields(self):
                if 'required' in field_.metadata and isinstance(getattr(self, field_.name), Empty):
                    missing.append(field_.name)
            if len(missing) > 0:
                raise TypeError(f'__init__ missing {len(missing)} required argument(s): {", ".join(missing)}')

    def __eq__(self, other):
        if not isinstance(other, self.__class__):
            return False

        return self.id_value() == other.id_value()

    def id_value(self):
        if not is_dataclass(self):
            raise TypeError('Entity::id_value() called on a non-dataclass entity')

        ret = []
        for field_ in fields(self):
            if 'id' in field_.metadata:
                ret.append(getattr(self, field_.name))

        return ret[0] if len(ret) == 1 else ret

    @classmethod
    def id_name(cls):
        if not is_dataclass(cls):
            raise TypeError('Entity::id_name() called on a non-dataclass entity')

        ret = []
        for field_ in fields(cls):
            if 'id' in field_.metadata:
                ret.append(field_.name)

        return ret[0] if len(ret) == 1 else ret

    @classmethod
    def match_id_from_argument_list(cls, args: dict):
        snake = f'{inflection.underscore(cls.__name__)}_id'
        if snake in args:
            return {snake: args[snake]}

        id_name = cls.id_name()
        if id_name in args:
            return {id_name: args[id_name]}

    def debug(self, *args, **kwargs):
        return self._logger.debug(*args, **kwargs)

    def info(self, *args, **kwargs):
        return self._logger.info(*args, **kwargs)

    def warning(self, *args, **kwargs):
        return self._logger.warning(*args, **kwargs)

    def error(self, *args, **kwargs):
        return self._logger.error(*args, **kwargs)


def id_(is_uuid: bool = True, **kwargs):
    metadata = {'id': True, 'required': True, 'type': str}
    if is_uuid:
        metadata['length'] = 36
    metadata.update(kwargs)
    return field(default_factory=lambda: str(uuid.uuid4()), metadata=metadata) if is_uuid else \
        required(**metadata)


def list_(**kwargs):
    kwargs['type'] = list
    return field(default_factory=lambda: [], metadata=kwargs)


def dict_(**kwargs):
    kwargs['type'] = dict
    return field(default_factory=lambda: {}, metadata=kwargs)


def now(**kwargs):
    kwargs['type'] = datetime
    return field(default_factory=lambda: datetime.now(), metadata=kwargs)


def today(**kwargs):
    kwargs['type'] = date
    return field(default_factory=lambda: date.today(), metadata=kwargs)


def required(type_: type = None, **kwargs):
    if type_ is not None:
        kwargs['type'] = type_
    kwargs['required'] = True
    return field(default_factory=lambda: Empty(), metadata=kwargs)


def optional(type_: type = None, default=MISSING, **kwargs):
    if type_ is not None:
        kwargs['type'] = type_
    kwargs['required'] = False
    if not isinstance(default, MISSING.__class__):
        return field(default_factory=lambda: default, metadata=kwargs)
    return field(default=None, metadata=kwargs)


def hidden(**kwargs):
    kwargs['hidden'] = True
    return field(default=None, init=False, repr=False, compare=False, metadata=kwargs)



#  Copyright (c) 2019 JD Williams
#
#  This file is part of Firefly, a Python SOA framework built by JD Williams. Firefly is free software; you can
#  redistribute it and/or modify it under the terms of the GNU General Public License as published by the
#  Free Software Foundation; either version 3 of the License, or (at your option) any later version.
#
#  Firefly is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
#  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
#  Public License for more details. You should have received a copy of the GNU Lesser General Public
#  License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#  You should have received a copy of the GNU General Public License along with Firefly. If not, see
#  <http://www.gnu.org/licenses/>.

from __future__ import annotations

import json
import os
from typing import Optional

import pytest
import firefly as ff
import firefly.infrastructure as ffi


os.environ['FF_ENVIRONMENT'] = 'test'


@pytest.fixture(scope="session")
def config():
    raise Exception('You must provide a config fixture.')


class TestingKernel(ff.Kernel):
    def reset(self):
        pass


@pytest.fixture(scope="session")
def container(config):
    from firefly.application import Container
    Container.configuration = lambda self: ffi.MemoryConfigurationFactory()(config)
    Container.asyncio_message_transport = ffi.AsyncioMessageTransport
    Container.message_transport = ffi.FakeMessageTransport
    Container.mutex = ffi.MemoryMutex
    Container.kernel = TestingKernel

    Container.__annotations__['asyncio_message_transport'] = ffi.AsyncioMessageTransport
    Container.__annotations__['mutex'] = ffi.MemoryMutex

    c = Container()

    class Authenticator(ff.Handler):
        def handle(self, message: ff.Message) -> Optional[bool]:
            if hasattr(message, 'fail_authentication'):
                raise ff.UnauthenticatedError()

            if hasattr(message, 'scopes'):
                scopes = message.scopes
                if isinstance(scopes, str):
                    scopes = json.loads(scopes)
                message.headers['decoded_token'] = {
                    'scopes': scopes,
                }
            return True

    c.authenticator.add(Authenticator())

    c.kernel.boot()

    return c


@pytest.fixture(scope="session")
def kernel(container) -> ff.Kernel:
    return container.kernel


@pytest.fixture(scope="session")
def context_map(container) -> ff.ContextMap:
    return container.context_map


@pytest.fixture(scope="session")
def system_bus(container) -> ff.SystemBus:
    return container.system_bus


@pytest.fixture(scope="session")
def message_factory(container) -> ff.MessageFactory:
    return container.message_factory


@pytest.fixture(scope="session")
def serializer(container):
    return container.serializer


@pytest.fixture(scope="function")
def registry(container, request) -> ff.Registry:
    registry = container.registry

    for context in container.context_map.contexts:
        for entity in context.entities:
            if issubclass(entity, ff.AggregateRoot) and entity is not ff.AggregateRoot:
                try:
                    repository = registry(entity)
                    if isinstance(repository, ffi.RdbRepository):
                        repository.migrate_schema()
                except ff.FrameworkError:
                    pass

    def teardown():
        for context in container.context_map.contexts:
            for entity in context.entities:
                if issubclass(entity, ff.AggregateRoot) and entity is not ff.AggregateRoot:
                    try:
                        registry(entity).destroy()
                    except ff.FrameworkError:
                        pass

    request.addfinalizer(teardown)
    registry.clear_cache()
    return registry


@pytest.fixture(scope="session")
def transport(container):
    return container.message_transport


@pytest.fixture(scope="function")
async def client(container, system_bus, loop, aiohttp_client):
    container.web_server.loop = loop
    deployment = ff.Deployment(project='Test', environment='testing', provider='default')
    system_bus.dispatch('firefly.DeploymentCreated', {'deployment': deployment})
    agent = container.agent_factory('default')
    agent(deployment, start_server=False)
    container.web_server.initialize()
    container.transaction_handler.reset_level()
    return await aiohttp_client(container.web_server.app)



#  Copyright (c) 2019 JD Williams
#
#  This file is part of Firefly, a Python SOA framework built by JD Williams. Firefly is free software; you can
#  redistribute it and/or modify it under the terms of the GNU General Public License as published by the
#  Free Software Foundation; either version 3 of the License, or (at your option) any later version.
#
#  Firefly is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
#  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
#  Public License for more details. You should have received a copy of the GNU Lesser General Public
#  License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#  You should have received a copy of the GNU General Public License along with Firefly. If not, see
#  <http://www.gnu.org/licenses/>.

from abc import ABC, abstractmethod


class Logger(ABC):
    @abstractmethod
    def debug(self, *args, **kwargs):
        pass

    @abstractmethod
    def info(self, *args, **kwargs):
        pass

    @abstractmethod
    def warning(self, *args, **kwargs):
        pass

    @abstractmethod
    def error(self, *args, **kwargs):
        pass

    @abstractmethod
    def exception(self, *args, **kwargs):
        pass

    @abstractmethod
    def set_level_to_fatal(self):
        pass

    @abstractmethod
    def set_level_to_error(self):
        pass

    @abstractmethod
    def set_level_to_warning(self):
        pass

    @abstractmethod
    def set_level_to_info(self):
        pass

    @abstractmethod
    def set_level_to_debug(self):
        pass

    @abstractmethod
    def disable(self):
        pass

    @abstractmethod
    def get_level(self):
        pass

    @abstractmethod
    def set_level(self, level: any):
        pass


class LoggerAware:
    _logger: Logger = None

    def debug(self, *args, **kwargs):
        return self._logger.debug(*args, **kwargs)

    def info(self, *args, **kwargs):
        return self._logger.info(*args, **kwargs)

    def warning(self, *args, **kwargs):
        return self._logger.warning(*args, **kwargs)

    def error(self, *args, **kwargs):
        return self._logger.error(*args, **kwargs)

    def exception(self, *args, **kwargs):
        return self._logger.exception(*args, **kwargs)



#  Copyright (c) 2019 JD Williams
#
#  This file is part of Firefly, a Python SOA framework built by JD Williams. Firefly is free software; you can
#  redistribute it and/or modify it under the terms of the GNU General Public License as published by the
#  Free Software Foundation; either version 3 of the License, or (at your option) any later version.
#
#  Firefly is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
#  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
#  Public License for more details. You should have received a copy of the GNU Lesser General Public
#  License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#  You should have received a copy of the GNU General Public License along with Firefly. If not, see
#  <http://www.gnu.org/licenses/>.

from __future__ import annotations

from typing import List, Callable

from firefly.domain import error


class RegisterMiddleware:
    def __call__(self, index: int = None, buses: List[str] = None, cb: Callable = None, replace: type = None):
        """
        Decorator to mark a class for insertion into the middleware stack in at least one of the buses.

        :param index: Optional positional index at which to insert the middleware.
        :param buses: Optional list of buses to which to add this middleware. Can contain 'event', 'command' and/or
        'query'
        :param cb: Optional callback that takes the bus type and the current list of middleware and returns a numeric
        index at which to insert the middleware.
        :param replace: Optional class type. If provided, the first matching middleware in the bus will be replaced.
        :return:
        """
        def middleware_wrapper(cls):
            try:
                cls.set_middleware_config({
                    'index': index,
                    'buses': buses,
                    'cb': cb,
                    'replace': replace,
                })
            except AttributeError:
                raise error.FrameworkError('@middleware used on invalid target')
            return cls

        return middleware_wrapper


register_middleware = RegisterMiddleware()



#  Copyright (c) 2019 JD Williams
#
#  This file is part of Firefly, a Python SOA framework built by JD Williams. Firefly is free software; you can
#  redistribute it and/or modify it under the terms of the GNU General Public License as published by the
#  Free Software Foundation; either version 3 of the License, or (at your option) any later version.
#
#  Firefly is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
#  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
#  Public License for more details. You should have received a copy of the GNU Lesser General Public
#  License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#  You should have received a copy of the GNU General Public License along with Firefly. If not, see
#  <http://www.gnu.org/licenses/>.

from __future__ import annotations

import hashlib
from typing import Any, Callable

import firefly.domain as ffd
from firefly import Query, Command, Event
from .asyncio_message_transport import AsyncioMessageTransport


class FakeMessageTransport(ffd.MessageTransport, ffd.LoggerAware):
    _serializer: ffd.Serializer = None
    _asyncio_message_transport: AsyncioMessageTransport = None

    def __init__(self):
        self._responses = {}

    def register_handler(self, message: str, handler: Callable):
        self._responses[message] = handler

    def dispatch(self, event: Event) -> None:
        if str(event) not in self._responses:
            return self._asyncio_message_transport.dispatch(event)
        return self._responses[str(event)](event)

    def invoke(self, command: Command) -> Any:
        if str(command) not in self._responses:
            return self._asyncio_message_transport.invoke(command)
        return self._responses[str(command)](command)

    def request(self, query: Query) -> Any:
        if str(query) not in self._responses:
            return self._asyncio_message_transport.request(query)
        return self._responses[str(query)](query)



#  Copyright (c) 2019 JD Williams
#
#  This file is part of Firefly, a Python SOA framework built by JD Williams. Firefly is free software; you can
#  redistribute it and/or modify it under the terms of the GNU General Public License as published by the
#  Free Software Foundation; either version 3 of the License, or (at your option) any later version.
#
#  Firefly is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
#  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
#  Public License for more details. You should have received a copy of the GNU Lesser General Public
#  License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#  You should have received a copy of the GNU General Public License along with Firefly. If not, see
#  <http://www.gnu.org/licenses/>.

from __future__ import annotations

from typing import Union

# __pragma__('skip')
from abc import ABC
# __pragma__('noskip')
# __pragma__ ('ecom')
"""?
from firefly.presentation.web.polyfills import ABC
?"""
# __pragma__ ('noecom')

import firefly.domain as ffd

from firefly.domain.service.messaging.message_bus import MessageBus


class CommandBus(MessageBus):
    _message_factory: ffd.MessageFactory = None

    def invoke(self, command: Union[ffd.Command, str], data: dict = None, async_: bool = False):
        if isinstance(command, str):
            data = data or {}
            data['_async'] = async_
            command = self._message_factory.command(command, data)
        else:
            setattr(command, '_async', async_)

        return self.dispatch(command)


class CommandBusAware(ABC):
    _command_bus: CommandBus = None

    def invoke(self, command: Union[ffd.Command, str], data: dict = None, async_: bool = False):
        return self._command_bus.invoke(command, data, async_=async_)



#  Copyright (c) 2019 JD Williams
#
#  This file is part of Firefly, a Python SOA framework built by JD Williams. Firefly is free software; you can
#  redistribute it and/or modify it under the terms of the GNU General Public License as published by the
#  Free Software Foundation; either version 3 of the License, or (at your option) any later version.
#
#  Firefly is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
#  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
#  Public License for more details. You should have received a copy of the GNU Lesser General Public
#  License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#  You should have received a copy of the GNU General Public License along with Firefly. If not, see
#  <http://www.gnu.org/licenses/>.
#
#  This file is part of Firefly, a Python SOA framework built by JD Williams. Firefly is free software; you can
#  redistribute it and/or modify it under the terms of the GNU General Public License as published by the
#  Free Software Foundation; either version 3 of the License, or (at your option) any later version.
#
#  Firefly is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
#  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
#  Public License for more details. You should have received a copy of the GNU Lesser General Public
#  License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#  You should have received a copy of the GNU General Public License along with Firefly. If not, see
#  <http://www.gnu.org/licenses/>.
#
#  This file is part of Firefly, a Python SOA framework built by JD Williams. Firefly is free software; you can
#  redistribute it and/or modify it under the terms of the GNU General Public License as published by the
#  Free Software Foundation; either version 3 of the License, or (at your option) any later version.
#
#  Firefly is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
#  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
#  Public License for more details. You should have received a copy of the GNU Lesser General Public
#  License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#  You should have received a copy of the GNU General Public License along with Firefly. If not, see
#  <http://www.gnu.org/licenses/>.

from __future__ import annotations

import firefly as ff
import firefly.domain.error as error
from firefly.domain.error import ConfigurationError


class Timer:
    def __call__(self, cron: str = None, rate: str = None, environment: str = None):
        # if cron is None and rate is None:
        #     raise ConfigurationError('Timer must have a cron or rate')

        # rate is currently unused
        if cron is None:
            raise ConfigurationError('Timer must have a cron expression')

        parts = cron.split(' ')
        if len(parts) == 5:
            parts.append('*')
        if parts[2] != '?':
            parts[4] = '?'
        elif parts[4] != '?':
            parts[2] = '?'
        cron = ' '.join(parts)

        def wrapper(cls):
            t = ff.Timer(cron=cron, environment=environment, command=cls)
            try:
                cls.set_timer(t)
            except AttributeError:
                raise error.FrameworkError('@timer used on invalid target')
            return cls

        return wrapper


timer = Timer()



#  Copyright (c) 2019 JD Williams
#
#  This file is part of Firefly, a Python SOA framework built by JD Williams. Firefly is free software; you can
#  redistribute it and/or modify it under the terms of the GNU General Public License as published by the
#  Free Software Foundation; either version 3 of the License, or (at your option) any later version.
#
#  Firefly is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
#  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
#  Public License for more details. You should have received a copy of the GNU Lesser General Public
#  License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#  You should have received a copy of the GNU General Public License along with Firefly. If not, see
#  <http://www.gnu.org/licenses/>.

from __future__ import annotations

import importlib

import firefly.domain as ffd
import firefly_di as di


class LoadContainers(ffd.ApplicationService):
    _context_map: ffd.ContextMap = None
    _container: di.Container = None

    def __call__(self):
        self.debug('LoadContainers called')
        for context in self._context_map.contexts:
            self.debug(f'Checking context "{context.name}"')
            if context.name == 'firefly':
                context.container = self._container
            else:
                context.container = self._load_module(context.name, context.config)
                self.debug(f'Loaded container: {context.container}')
                self._container.register_container(context.container)
            self.dispatch(ffd.ContainerInitialized(context=context.name))

        self.debug('All containers are built. Looping through again to link them all together.')
        for context in self._context_map.contexts:
            for name, config in (context.config.get('extensions', {}) or {}).items():
                self.debug(f'Registering {name} container with {context.name}')
                context.container.register_container(self._context_map.get_context(name).container)
            self.debug('Registering root container')
            context.container.register_container(self._container)

            if context.config.get('extends'):
                for c in self._context_map.contexts:
                    if c.name != context.config.get('extends'):
                        self._context_map.get_context(context.config.get('extends')).container.register_container(
                            c.container
                        )

        self.dispatch(ffd.ContainersLoaded())

    def _load_module(self, name: str, config: dict) -> di.Container:
        module_name = config.get('container_module', '{}.application')
        try:
            module = importlib.import_module(module_name.format(name))
            container_cls = getattr(module, config.get('container_name', 'Container'))
        except (ModuleNotFoundError, AttributeError):
            class EmptyContainer(di.Container):
                pass
            container_cls = EmptyContainer

        container = container_cls()

        return container



#  Copyright (c) 2019 JD Williams
#
#  This file is part of Firefly, a Python SOA framework built by JD Williams. Firefly is free software; you can
#  redistribute it and/or modify it under the terms of the GNU General Public License as published by the
#  Free Software Foundation; either version 3 of the License, or (at your option) any later version.
#
#  Firefly is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
#  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
#  Public License for more details. You should have received a copy of the GNU Lesser General Public
#  License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#  You should have received a copy of the GNU General Public License along with Firefly. If not, see
#  <http://www.gnu.org/licenses/>.
#
#  This file is part of Firefly, a Python SOA framework built by JD Williams. Firefly is free software; you can
#  redistribute it and/or modify it under the terms of the GNU General Public License as published by the
#  Free Software Foundation; either version 3 of the License, or (at your option) any later version.
#
#  Firefly is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
#  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
#  Public License for more details. You should have received a copy of the GNU Lesser General Public
#  License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#  You should have received a copy of the GNU General Public License along with Firefly. If not, see
#  <http://www.gnu.org/licenses/>.

from __future__ import annotations

import re

# __pragma__('skip')
from abc import ABC, abstractmethod
from dateutil.parser import parse
from typing import Any
# __pragma__('noskip')
# __pragma__ ('ecom')
"""?
from firefly.presentation.web.polyfills import ABC, abstractmethod, Any
?"""
# __pragma__ ('noecom')


class Validation(ABC):
    def __init__(self, message: str = None):
        self.message = message

    @abstractmethod
    def __call__(self, value: Any, dto: dict) -> bool:
        pass


class IsType(Validation):
    def __init__(self, type_: Any, message: str = None):
        super().__init__(message or '{} must be of type ' + type_)
        self._type = type_

    def __call__(self, value: Any, dto: dict) -> bool:
        return isinstance(value, self._type)


class HasLength(Validation):
    def __init__(self, length: int, message: str = None):
        super().__init__(message or '{} must have a length of ' + str(length))
        self.length = length

    def __call__(self, value: Any, dto: dict) -> bool:
        return len(value) == self.length


class HasMaxLength(Validation):
    def __init__(self, length: int, message: str = None):
        super().__init__(message or '{} must have a length of ' + str(length))
        self.length = length

    def __call__(self, value: Any, dto: dict) -> bool:
        return len(value) <= self.length


class HasMinLength(Validation):
    def __init__(self, length: int, message: str = None):
        super().__init__(message or '{} must have a length of ' + str(length))
        self.length = length

    def __call__(self, value: Any, dto: dict) -> bool:
        return len(value) >= self.length


class Matches(Validation):
    def __init__(self, other_field: str, message: str = None):
        self._other_field = other_field
        super().__init__(message or '{} must match ' + other_field)

    def __call__(self, value: Any, dto: dict) -> bool:
        return value == dto[self._other_field]


class MatchesPattern(Validation):
    def __init__(self, pattern: str, message: str = None):
        super().__init__(message or '{} must match the pattern ' + pattern)
        self.regex = re.compile(pattern)

    def __call__(self, value: Any, dto: dict) -> bool:
        return self.regex.fullmatch(value) is not None


class IsValidEmail(MatchesPattern):
    def __init__(self, message: str = None):
        super().__init__(
            r"(^[a-zA-Z0-9_.+-]+@[a-zA-Z0-9-]+\.[a-zA-Z0-9-.]+$)",
            message or '{} must be a valid email address'
        )


class IsValidUrl(MatchesPattern):
    def __init__(self, message: str = None):
        super().__init__(
            r"http[s]?://(?:[a-zA-Z]|[0-9]|[$-_@.&+]|[!*\(\),]|(?:%[0-9a-fA-F][0-9a-fA-F]))+",
            message or '{} must be a valid url'
        )


class IsInt(MatchesPattern):
    def __init__(self, message: str = None):
        super().__init__(r'^\d+$', message or '{} must be an integer')

    def __call__(self, value: Any, dto: dict) -> bool:
        return isinstance(value, int) or (isinstance(value, str) and self.regex.fullmatch(value) is not None)


class IsFloat(MatchesPattern):
    def __init__(self, message: str = None):
        super().__init__(r'^\d+\.\d+$', message or '{} must be a float')

    def __call__(self, value: Any, dto: dict) -> bool:
        return isinstance(value, float) or (isinstance(value, str) and self.regex.fullmatch(value) is not None)


class IsNumeric(MatchesPattern):
    def __init__(self, message: str = None):
        super().__init__(r'^\d+\.?\d*$', message or '{} must be numeric')

    def __call__(self, value: Any, dto: dict) -> bool:
        return isinstance(value, (int, float)) or (isinstance(value, str) and self.regex.fullmatch(value) is not None)


class NumericValidation(Validation, ABC):
    def __init__(self, value: float, message: str = None):
        super().__init__(message)
        self.value = value

    @abstractmethod
    def __call__(self, value: Any, dto: dict) -> bool:
        pass


class IsMultipleOf(NumericValidation):
    def __call__(self, value: Any, dto: dict) -> bool:
        return value % self.value == 0


class IsLessThan(NumericValidation):
    def __call__(self, value: Any, dto: dict) -> bool:
        return value < self.value


class IsLessThanOrEqualTo(NumericValidation):
    def __call__(self, value: Any, dto: dict) -> bool:
        return value <= self.value


class IsGreaterThan(NumericValidation):
    def __call__(self, value: Any, dto: dict) -> bool:
        return value > self.value


class IsGreaterThanOrEqualTo(NumericValidation):
    def __call__(self, value: Any, dto: dict) -> bool:
        return value >= self.value


class IsDatetime(Validation):
    def __init__(self, message: str = None):
        super().__init__(message)

    def __call__(self, value: Any, dto: dict) -> bool:
        # __pragma__ ('ecom')
        """?
        return moment(value).isValid();
        ?"""
        # __pragma__ ('noecom')
        # __pragma__('skip')
        try:
            parse(value)
            return True
        except (ValueError, OverflowError):
            return False
        # __pragma__('noskip')


class IsOneOf(Validation):
    def __init__(self, values, message: str = None):
        super().__init__(message or '{} must be one of ' + str(values))
        self.values = values

    def __call__(self, value: Any, dto: dict) -> bool:
        values = self.values
        if callable(values):
            values = values(dto)
        return value in values



#  Copyright (c) 2019 JD Williams
#
#  This file is part of Firefly, a Python SOA framework built by JD Williams. Firefly is free software; you can
#  redistribute it and/or modify it under the terms of the GNU General Public License as published by the
#  Free Software Foundation; either version 3 of the License, or (at your option) any later version.
#
#  Firefly is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
#  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
#  Public License for more details. You should have received a copy of the GNU Lesser General Public
#  License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#  You should have received a copy of the GNU General Public License along with Firefly. If not, see
#  <http://www.gnu.org/licenses/>.

from __future__ import annotations

import importlib
from abc import ABC
import firefly_di as di
import firefly.domain as ffd


class ModuleBase(ABC, ffd.LoggerAware):
    def load_container(self, name: str, bus: ffd.MessageBus):
        try:
            self.debug('Attempting to import module {}.application'.format(name))
            module = importlib.import_module('{}.application'.format(name))
            container_class = getattr(module, 'Container')
            self.debug('Container imported successfully')
        except (ModuleNotFoundError, AttributeError):
            self.debug('Failed to load application module for {}. Ignoring.', name)

            class EmptyContainer(di.Container):
                pass
            container_class = EmptyContainer

        self.debug('Injecting message bus into container')
        container_class.message_bus = bus
        container_class.__annotations__['message_bus'] = ffd.MessageBus
        container = container_class()

        return container



#  Copyright (c) 2019 JD Williams
#
#  This file is part of Firefly, a Python SOA framework built by JD Williams. Firefly is free software; you can
#  redistribute it and/or modify it under the terms of the GNU General Public License as published by the
#  Free Software Foundation; either version 3 of the License, or (at your option) any later version.
#
#  Firefly is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
#  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
#  Public License for more details. You should have received a copy of the GNU Lesser General Public
#  License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#  You should have received a copy of the GNU General Public License along with Firefly. If not, see
#  <http://www.gnu.org/licenses/>.

from __future__ import annotations

from typing import List

import firefly as ff

from .task import Task
from .user import User


@ff.rest.crud()
class TodoList(ff.AggregateRoot, create_on='iam.UserCreated', delete_on='iam.UserDeleted'):
    id: str = ff.id_()
    user: User = ff.required()
    name: str = ff.optional(index=True, length=128)
    tasks: List[Task] = ff.list_()

    def __post_init__(self):
        if self.name is None:
            self.name = f"{self.user.name}'s TODO List"

    @ff.rest('/task', method='POST')
    def add_task(self, task: Task) -> ff.EventList:
        self.tasks.append(task)
        return 'TaskAdded', task

    def remove_task(self, task: Task):
        self.tasks.remove(task)

    def complete_task(self, task_id: str) -> ff.EventList:
        for task in self.tasks:
            if task_id == task.id:
                task.complete_task()
                return 'TaskCompleted', task
        raise Exception(f'Task {task_id} not found in TodoList {self}')



#  Copyright (c) 2019 JD Williams
#
#  This file is part of Firefly, a Python SOA framework built by JD Williams. Firefly is free software; you can
#  redistribute it and/or modify it under the terms of the GNU General Public License as published by the
#  Free Software Foundation; either version 3 of the License, or (at your option) any later version.
#
#  Firefly is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
#  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
#  Public License for more details. You should have received a copy of the GNU Lesser General Public
#  License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#  You should have received a copy of the GNU General Public License along with Firefly. If not, see
#  <http://www.gnu.org/licenses/>.
#
#  This file is part of Firefly, a Python SOA framework built by JD Williams. Firefly is free software; you can
#  redistribute it and/or modify it under the terms of the GNU General Public License as published by the
#  Free Software Foundation; either version 3 of the License, or (at your option) any later version.
#
#  Firefly is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
#  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
#  Public License for more details. You should have received a copy of the GNU Lesser General Public
#  License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#  You should have received a copy of the GNU General Public License along with Firefly. If not, see
#  <http://www.gnu.org/licenses/>.

from __future__ import annotations

from multiprocessing import Pipe
from multiprocessing.context import Process
from typing import Callable, Iterable

from ..core.domain_service import DomainService


def wrapper(cb):
    def _wrapper(*args):
        conn = args[-1]
        try:
            conn.send(cb(*args[0:-1]))
            conn.close()
        except Exception as e:
            conn.send(e)
        finally:
            conn.close()
    return _wrapper


class BatchProcess(DomainService):
    def __call__(self, cb: Callable, args: Iterable, timeout: int = None):
        processes = []
        connections = []

        for arg_set in args:
            parent, child = Pipe(duplex=False)
            connections.append(parent)
            process = Process(target=wrapper(cb), args=arg_set + (child,))
            processes.append(process)

        for p in processes:
            p.start()

        ret = []
        for c in connections:
            ret.append(c.recv())

        for p in processes:
            p.join(timeout=timeout)

        return ret



#  Copyright (c) 2019 JD Williams
#
#  This file is part of Firefly, a Python SOA framework built by JD Williams. Firefly is free software; you can
#  redistribute it and/or modify it under the terms of the GNU General Public License as published by the
#  Free Software Foundation; either version 3 of the License, or (at your option) any later version.
#
#  Firefly is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
#  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
#  Public License for more details. You should have received a copy of the GNU Lesser General Public
#  License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#  You should have received a copy of the GNU General Public License along with Firefly. If not, see
#  <http://www.gnu.org/licenses/>.

from __future__ import annotations

import importlib
import os
from dataclasses import is_dataclass
from typing import Any

import firefly.domain as ffd
from firefly import Message
from jinja2 import Environment, TemplateNotFound, BaseLoader, select_autoescape


class HtmlConverter(ffd.ContentConverter):
    def __init__(self, config: ffd.Configuration):
        self._environments = {}

        for name, context in config.contexts.items():
            context = context or {}
            self._environments[name] = Environment(
                loader=TemplateLoader(name, context),
                autoescape=select_autoescape(
                    context.get('output', {}).get('autoescape', ['html', 'xml'])
                )
            )

    def can_convert(self, request: Message, response: Any):
        try:
            self._get_template(request)
            return True
        except (KeyError, TemplateNotFound):
            return False

    def convert(self, request: Message, response: Any):
        template = self._get_template(request)

        body = None
        if isinstance(response, dict) and not is_dataclass(response):
            body = template.render(**response)
        elif isinstance(response, Message):
            body = template.render(**response.to_dict())

        if body is not None:
            response = ffd.HttpResponse(body=body, headers={
                'Content-Type': 'text/html',
            })

        return response

    def _get_template(self, message: Message):
        request = message.headers['http_request']
        return self._environments[message.get_context()]\
            .get_template(f"{request['path']}.{request['method'].lower()}.html")


class TemplateLoader(BaseLoader):
    def __init__(self, context: str, config: dict):
        self._paths = []
        try:
            m = importlib.import_module('templates')
            for path in m.__path__:
                self._paths.append(path)
        except ModuleNotFoundError:
            pass

        module_name = config.get('template_module', f'{context}.ui.templates')
        try:
            m = importlib.import_module(module_name)
            for path in m.__path__:
                self._paths.append(path)
        except ModuleNotFoundError:
            pass

    def get_source(self, environment, template):
        for path in self._paths:
            full_path = os.path.join(path, str(template).lstrip('/'))
            if not os.path.exists(full_path):
                raise TemplateNotFound(template)
            mtime = os.path.getmtime(full_path)
            with open(full_path, 'r') as fp:
                source = fp.read()
            return source, full_path, lambda: mtime == os.path.getmtime(full_path)

        raise TemplateNotFound(template)



#  Copyright (c) 2019 JD Williams
#
#  This file is part of Firefly, a Python SOA framework built by JD Williams. Firefly is free software; you can
#  redistribute it and/or modify it under the terms of the GNU General Public License as published by the
#  Free Software Foundation; either version 3 of the License, or (at your option) any later version.
#
#  Firefly is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
#  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
#  Public License for more details. You should have received a copy of the GNU Lesser General Public
#  License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#  You should have received a copy of the GNU General Public License along with Firefly. If not, see
#  <http://www.gnu.org/licenses/>.

from __future__ import annotations

import inspect
import json
from dataclasses import fields, is_dataclass
from datetime import datetime, date
from typing import Type, List

import firefly.domain as ffd
import inflection
import typing


@ffd.cli('firefly generate api-spec')
class GenerateApiSpec(ffd.ApplicationService):
    _context_map: ffd.ContextMap = None

    def __init__(self):
        self._schemas = []

    def __call__(self, tag: str = None, **kwargs):
        try:
            from apispec import APISpec
            from docstring_parser import parse
        except ModuleNotFoundError:
            raise ffd.MissingDependency('This feature requires "apispec" and "docstring_parser"')

        self._spec = APISpec(
            title='Testing Spec',
            version='1.0.0',
            openapi_version='3.0.2',
        )

        for context in self._context_map.contexts:
            prefix = f'/{inflection.dasherize(context.name)}'
            for endpoint in context.endpoints:
                if isinstance(endpoint, ffd.HttpEndpoint):
                    if tag is not None and tag not in endpoint.tags:
                        continue
                    docstring = parse(endpoint.service.__call__.__doc__)
                    short_description = docstring.short_description \
                        if docstring.short_description != 'Call self as a function.' else None

                    path = endpoint.route
                    if not path.startswith(prefix):
                        path = f'{prefix}{path}'

                    request_body = self._request_body(endpoint)
                    method = endpoint.method.lower()
                    operations = {method: {}}
                    if request_body is not None:
                        operations[method]['requestBody'] = request_body
                    self._spec.path(
                        path=path,
                        operations=operations,
                        parameters=self._parameter_list(endpoint)
                    )

        print(json.dumps(self._spec.to_dict(), indent=2))

    def _parameter_list(self, endpoint: ffd.HttpEndpoint):
        ret = []
        if inspect.isclass(endpoint.service) and issubclass(endpoint.service, ffd.AggregateRoot):
            if endpoint.method.lower() == 'post':
                return [
                    {
                        'name': field_.name,
                        'in': 'TODO',
                        'description': 'TODO',
                        'required': 'required' in field_.metadata and field_.metadata['required'] is True,
                        'schema': {
                            'type': field_.type,
                        }
                    }
                    for field_ in list(filter(lambda f: not f.name.startswith('_'), fields(endpoint.service)))
                ]

    def _request_body(self, endpoint: ffd.HttpEndpoint):
        if not inspect.isclass(endpoint.service):
            return None

        if endpoint.method.lower() in ('post', 'put', 'patch'):
            return {
                'required': True,
                'content': {
                    'application/json': {'schema': {
                        '$ref': f'#/components/schemas/{self._add_schema(endpoint.service)}'}
                    }
                }
            }

    def _add_schema(self, cls: typing.Union[Type[ffd.ValueObject], Type[ffd.ApplicationService]]):
        try:
            if cls.__name__ in self._schemas:
                return cls.__name__
        except AttributeError:
            pass

        if issubclass(cls, ffd.ValueObject):
            return self._add_entity_schema(cls)
        elif issubclass(cls, ffd.ApplicationService):
            return self._add_service_body_schema(cls)

    def _add_service_body_schema(self, cls: Type[ffd.ApplicationService]):
        props = {}
        signature = inspect.signature(cls.__call__)
        hints = typing.get_type_hints(cls.__call__)
        for name, param in signature.parameters.items():
            if name == 'self' or param.kind in (inspect.Parameter.VAR_POSITIONAL, inspect.Parameter.VAR_KEYWORD):
                continue
            try:
                type_ = hints[name]
            except KeyError:
                type_ = str
            self._handle_type(name, type_, props)

        self._spec.components.schema(cls.__name__, {'properties': props})
        self._schemas.append(cls.__name__)
        return cls.__name__

    def _add_entity_schema(self, cls: Type[ffd.ValueObject]):
        props = {}
        hints = typing.get_type_hints(cls)
        if not is_dataclass(cls):
            return None
        for field_ in fields(cls):
            if field_.name.startswith('_'):
                continue
            type_ = hints[field_.name]
            self._handle_type(field_.name, type_, props)

        self._spec.components.schema(cls.__name__, {'properties': props})
        self._schemas.append(cls.__name__)
        return cls.__name__

    def _handle_type(self, name: str, type_: type, props: dict):
        if inspect.isclass(type_) and issubclass(type_, (ffd.Entity, ffd.ValueObject)):
            props[name] = {
                'type': {
                    'schema': {
                        '$ref': f'#/components/schemas/{self._add_schema(type_)}'
                    }
                }
            }
        elif ffd.is_type_hint(type_) and ffd.get_origin(type_) is typing.List:
            t = ffd.get_args(type_)[0]
            props[name] = {
                'type': 'array',
                'items': {
                    '$ref': f'#/components/schemas/{self._add_schema(t)}'
                }
            }
        elif ffd.is_type_hint(type_) and ffd.get_origin(type_) is typing.Union:
            args = ffd.get_args(type_)
            if args[1] is None:
                props[name] = self._map_type(args[0])
                props[name].update({'required': False})
        elif ffd.is_type_hint(type_) and ffd.get_origin(type_) is typing.Dict:
            props[name] = {'type': 'object'}
        else:
            props[name] = self._map_type(type_)

    def _map_type(self, type_: type):
        if type_ is str or type_ == 'str':
            return {'type': 'string'}
        elif type_ is int or type_ == 'int':
            return {'type': 'integer'}
        elif type_ is float or type_ == 'float':
            return {'type': 'number'}
        elif type_ is bool or type_ == 'bool':
            return {'type': 'boolean'}
        elif type_ is list or type_ == 'list':
            return {'type': 'array'}
        elif type_ is dict or type_ == 'dict':
            return {'type': 'object'}
        elif type_ is datetime or type_ == 'datetime':
            return {'type': 'string', 'format': 'date-time'}
        elif type_ is date or type_ == 'date':
            return {'type': 'string', 'format': 'date'}

        raise Exception(f"Don't know how to handle type: {type_}")



#  Copyright (c) 2019 JD Williams
#
#  This file is part of Firefly, a Python SOA framework built by JD Williams. Firefly is free software; you can
#  redistribute it and/or modify it under the terms of the GNU General Public License as published by the
#  Free Software Foundation; either version 3 of the License, or (at your option) any later version.
#
#  Firefly is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
#  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
#  Public License for more details. You should have received a copy of the GNU Lesser General Public
#  License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#  You should have received a copy of the GNU General Public License along with Firefly. If not, see
#  <http://www.gnu.org/licenses/>.

from __future__ import annotations

from functools import reduce
from typing import List, Callable, Optional, Union

import firefly.domain as ffd
from firefly.domain.repository.repository import T


class MemoryRepository(ffd.Repository[T]):
    def __init__(self):
        super().__init__()
        self.entities = []
        self._index = 0

    def append(self, entity: T):
        self.entities.append(entity)

    def remove(self, entity: T):
        self.entities.remove(entity)

    def find(self, exp: Union[str, Callable]) -> Optional[T]:
        if isinstance(exp, str):
            for e in self.entities:
                if e.id_value() == exp:
                    return e
        else:
            criteria = self._get_search_criteria(exp)
            results = list(filter(lambda i: criteria.matches(i), self.entities))
            if len(results) > 1:
                raise ffd.MultipleResultsFound()
            if len(results) == 0:
                raise ffd.NoResultFound()
            return results[0]

    def filter(self, cb: Callable) -> List[T]:
        criteria = self._get_search_criteria(cb)
        return list(filter(lambda i: criteria.matches(i), self.entities))

    def reduce(self, cb: Callable) -> Optional[T]:
        return reduce(cb, self.entities)

    def __iter__(self):
        return self

    def __next__(self):
        if self._index >= len(self.entities):
            self._index = 0
            raise StopIteration()
        self._index += 1
        return self.entities[self._index - 1]

    def __len__(self):
        return len(self.entities)

    def __getitem__(self, item):
        return self.entities[item]

    def commit(self):
        pass



#  Copyright (c) 2019 JD Williams
#
#  This file is part of Firefly, a Python SOA framework built by JD Williams. Firefly is free software; you can
#  redistribute it and/or modify it under the terms of the GNU General Public License as published by the
#  Free Software Foundation; either version 3 of the License, or (at your option) any later version.
#
#  Firefly is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
#  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
#  Public License for more details. You should have received a copy of the GNU Lesser General Public
#  License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#  You should have received a copy of the GNU General Public License along with Firefly. If not, see
#  <http://www.gnu.org/licenses/>.
#
#  This file is part of Firefly, a Python SOA framework built by JD Williams. Firefly is free software; you can
#  redistribute it and/or modify it under the terms of the GNU General Public License as published by the
#  Free Software Foundation; either version 3 of the License, or (at your option) any later version.
#
#  Firefly is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
#  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
#  Public License for more details. You should have received a copy of the GNU Lesser General Public
#  License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#  You should have received a copy of the GNU General Public License along with Firefly. If not, see
#  <http://www.gnu.org/licenses/>.

from __future__ import annotations

import firefly.domain as ff
import firefly.infrastructure as ffi


@ff.cli('firefly migrate-repositories')
class MigrateRepositories(ff.ApplicationService):
    _registry: ff.Registry = None
    _context_map: ff.ContextMap = None

    def __call__(self, **kwargs):
        for context in self._context_map.contexts:
            if context.is_extension:
                continue

            for entity in context.entities:
                if not issubclass(entity, ff.AggregateRoot):
                    continue

                try:
                    repository = self._registry(entity)
                except ff.FrameworkError as e:
                    if 'No registry found' not in str(e):
                        raise e
                    else:
                        continue

                if isinstance(repository, ffi.RdbRepository):
                    print(f'Migrating {repository.__class__.__name__}')
                    repository.migrate_schema()



#  Copyright (c) 2019 JD Williams
#
#  This file is part of Firefly, a Python SOA framework built by JD Williams. Firefly is free software; you can
#  redistribute it and/or modify it under the terms of the GNU General Public License as published by the
#  Free Software Foundation; either version 3 of the License, or (at your option) any later version.
#
#  Firefly is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
#  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
#  Public License for more details. You should have received a copy of the GNU Lesser General Public
#  License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#  You should have received a copy of the GNU General Public License along with Firefly. If not, see
#  <http://www.gnu.org/licenses/>.

from __future__ import annotations

import importlib
import inspect

import firefly.domain as ffd
import inflection


class LoadDomainLayer(ffd.ApplicationService):
    _context_map: ffd.ContextMap = None
    _logger: ffd.Logger = None

    def __call__(self, **kwargs):
        for context in self._context_map.contexts:
            self._load_module(context)
            if 'extends' in context.config:
                base_context = self._context_map.get_context(context.config['extends'])
                self._load_module(
                    context, base_context.config.get('entity_module', '{}.domain').format(base_context.name)
                )
        self.dispatch(ffd.DomainEntitiesLoaded())

    def _load_module(self, context: ffd.Context, module_name: str = None):
        module_name = module_name or context.config.get('entity_module', '{}.domain').format(context.name)
        try:
            module = importlib.import_module(module_name)
        except (ModuleNotFoundError, KeyError):
            return

        for k, v in module.__dict__.items():
            if not inspect.isclass(v):
                continue
            if issubclass(v, ffd.Entity):
                v._logger = self._logger
                context.entities.append(v)
            elif issubclass(v, ffd.ValueObject):
                v._logger = self._logger
            elif issubclass(v, ffd.DomainService) and v is not ffd.DomainService:
                v._logger = self._logger
                v._system_bus = self._system_bus
                name = inflection.underscore(v.__name__)
                setattr(context.container.__class__, name, v)
                if not hasattr(context.container.__class__, '__annotations__'):
                    context.container.__class__.__annotations__ = {}
                context.container.__class__.__annotations__[name] = v
                context.container.clear_annotation_cache()



#!/usr/bin/env python
#  Copyright (c) 2019 JD Williams
#
#  This file is part of Firefly, a Python SOA framework built by JD Williams. Firefly is free software; you can
#  redistribute it and/or modify it under the terms of the GNU General Public License as published by the
#  Free Software Foundation; either version 3 of the License, or (at your option) any later version.
#
#  Firefly is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
#  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
#  Public License for more details. You should have received a copy of the GNU Lesser General Public
#  License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#  You should have received a copy of the GNU General Public License along with Firefly. If not, see
#  <http://www.gnu.org/licenses/>.

""" Unit tests for the msvs.py file. """

import gyp.generator.msvs as msvs
import unittest
import StringIO


class TestSequenceFunctions(unittest.TestCase):

  def setUp(self):
    self.stderr = StringIO.StringIO()

  def test_GetLibraries(self):
    self.assertEqual(
      msvs._GetLibraries({}),
      [])
    self.assertEqual(
      msvs._GetLibraries({'libraries': []}),
      [])
    self.assertEqual(
      msvs._GetLibraries({'other':'foo', 'libraries': ['a.lib']}),
      ['a.lib'])
    self.assertEqual(
      msvs._GetLibraries({'libraries': ['-la']}),
      ['a.lib'])
    self.assertEqual(
      msvs._GetLibraries({'libraries': ['a.lib', 'b.lib', 'c.lib', '-lb.lib',
                                   '-lb.lib', 'd.lib', 'a.lib']}),
      ['c.lib', 'b.lib', 'd.lib', 'a.lib'])

if __name__ == '__main__':
  unittest.main()



#  Copyright (c) 2019 JD Williams
#
#  This file is part of Firefly, a Python SOA framework built by JD Williams. Firefly is free software; you can
#  redistribute it and/or modify it under the terms of the GNU General Public License as published by the
#  Free Software Foundation; either version 3 of the License, or (at your option) any later version.
#
#  Firefly is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
#  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
#  Public License for more details. You should have received a copy of the GNU Lesser General Public
#  License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#  You should have received a copy of the GNU General Public License along with Firefly. If not, see
#  <http://www.gnu.org/licenses/>.

from __future__ import annotations

from dataclasses import is_dataclass
from datetime import datetime, date
from pprint import pprint
from typing import Union, List, Type

import firefly.domain as ffd
import regex

T = '__FF_SKIP_TYPE'


class EntityAttributeSpy:
    def __init__(self, type_: Type[ffd.ValueObject] = None):
        self._type = type_

    def __getattribute__(self, item):
        if object.__getattribute__(self, '_type') is not None:
            t = object.__getattribute__(self, '_type')
            valid = False
            if is_dataclass(t) and item in t.__dataclass_fields__:
                valid = True
            elif hasattr(t, item):
                valid = True
            if not valid:
                raise AttributeError(f"'{t.__name__}' object has no attribute '{item}'")
        return Attr(item)


class AttributeString:
    __modifiers = None
    _value: str = None

    def __init__(self, value: str):
        if '(' in value:
            matches = list(map(lambda rr: rr[2], regex.findall(r'((\w)\((?R)\))|(\w+)', str(value))))
            attribute = matches.pop()

            for func in reversed(matches):
                self.add_modifier(func)
            self._value = attribute
        else:
            self._value = str(value)

    def add_modifier(self, modifier: str):
        if self.__modifiers is None:
            self.__modifiers = []
        self.__modifiers.append(modifier)

    def has_modifiers(self):
        return self.__modifiers is not None

    def get_modifiers(self):
        return self.__modifiers

    def __str__(self):
        return str(self._value)

    def __repr__(self):
        ret = ''
        if self.has_modifiers():
            for modifier in self.get_modifiers():
                ret += f'{modifier}('
        ret += str(self)
        if self.has_modifiers():
            for _ in self.get_modifiers():
                ret += ')'

        return ret


class Invalid:
    pass


INVALID = Invalid()


class Attr:
    def __init__(self, attr: str, default=None):
        self.attr = AttributeString(attr)
        self.default = default

    def __getattribute__(self, item):
        if item in ('has_modifiers', 'get_modifiers'):
            return getattr(self.attr, item)

        try:
            return object.__getattribute__(self, item)
        except AttributeError:
            if not item.startswith('_'):
                self.attr = AttributeString(f'{self.attr}.{item}')
                return self
            raise ffd.InvalidOperand(f"Use of '{item}' is not currently supported.")

    def is_none(self):
        return BinaryOp(self.attr, 'is', 'null')

    def is_not_none(self):
        return BinaryOp(self.attr, 'is not', 'null')

    def is_false(self):
        return BinaryOp(self.attr, 'is', False)

    def is_true(self):
        return BinaryOp(self.attr, 'is', True)

    def is_in(self, set_):
        return BinaryOp(self.attr, 'in', set_)

    def not_in(self, set_):
        return BinaryOp(self.attr, 'not in', set_)

    def contains(self, value):
        return BinaryOp(self.attr, 'contains', value)

    def startswith(self, value):
        return BinaryOp(self.attr, 'startswith', value)

    def lower(self):
        self.attr.add_modifier('LOWER')
        return self

    def upper(self):
        self.attr.add_modifier('UPPER')
        return self

    def endswith(self, value):
        return BinaryOp(self.attr, 'endswith', value)

    def reversed(self):
        return self.attr, True

    def __eq__(self, other):
        return BinaryOp(self.attr, '==', other)

    def __ne__(self, other):
        return BinaryOp(self.attr, '!=', other)

    def __gt__(self, other):
        return BinaryOp(self.attr, '>', other)

    def __ge__(self, other):
        return BinaryOp(self.attr, '>=', other)

    def __lt__(self, other):
        return BinaryOp(self.attr, '<', other)

    def __le__(self, other):
        return BinaryOp(self.attr, '<=', other)

    def __repr__(self):
        return repr(self.attr)

    def __str__(self):
        return str(self.attr)


class AttrFactory:
    def __init__(self, fields: List[str]):
        self._fields = fields

    def __getattr__(self, item):
        if item in self._fields:
            return Attr(item)
        return object.__getattribute__(self, item)


class BinaryOp:
    def __init__(self, lhv, op, rhv):
        self.lhv = lhv
        self.op = op
        self.rhv = rhv

    def to_dict(self):
        return self._do_to_dict(self)

    def _do_to_dict(self, bop: BinaryOp):
        return {
            'l': self._get_serialized_value(bop.lhv),
            'o': bop.op,
            'r': self._get_serialized_value(bop.rhv)
        }

    def _get_serialized_value(self, val):
        if isinstance(val, BinaryOp):
            return self._do_to_dict(val)
        elif isinstance(val, (Attr, AttributeString)):
            return f'a:{repr(val)}'
        elif isinstance(val, datetime):
            return f'datetime:{val}'
        elif isinstance(val, date):
            return f'date:{val}'
        else:
            return val

    @classmethod
    def from_dict(cls, data: dict):
        return BinaryOp(
            cls._deserialize_value(data['l']),
            data['o'],
            cls._deserialize_value(data['r'])
        )

    @classmethod
    def _deserialize_value(cls, val):
        if isinstance(val, dict):
            return cls.from_dict(val)
        elif isinstance(val, str) and val.startswith('a:'):
            return Attr(val.split(':')[1])
        elif isinstance(val, str) and val.startswith('datetime:'):
            return ffd.parse(val.split(':')[1])
        elif isinstance(val, str) and val.startswith('date:'):
            return ffd.parse(val.split(':')[1]).date()
        else:
            return val

    def matches(self, data: Union[ffd.Entity, dict]) -> bool:
        if isinstance(data, ffd.Entity):
            data = data.to_dict(force_all=True)

        return self._do_match(self, data)

    def _do_match(self, bop: BinaryOp, data: dict) -> bool:
        if isinstance(bop.lhv, BinaryOp):
            lhv = self._do_match(bop.lhv, data)
        elif isinstance(bop.lhv, AttributeString):
            lhv = self._parse_attribute_string(bop.lhv, data)
        elif isinstance(bop.lhv, Attr):
            lhv = self._parse_attribute_string(bop.lhv.attr, data)
        else:
            lhv = bop.lhv

        if isinstance(bop.rhv, BinaryOp):
            rhv = self._do_match(bop.rhv, data)
        elif isinstance(bop.rhv, AttributeString):
            rhv = self._parse_attribute_string(bop.rhv, data)
        elif isinstance(bop.rhv, Attr):
            rhv = self._parse_attribute_string(bop.rhv.attr, data)
        else:
            rhv = bop.rhv

        if bop.op == '==':
            return lhv == rhv
        if bop.op == '!=':
            return lhv != rhv
        if bop.op == '>':
            return lhv > rhv
        if bop.op == '<':
            return lhv < rhv
        if bop.op == '>=':
            return lhv >= rhv
        if bop.op == '<=':
            return lhv <= rhv
        if bop.op == 'is':
            return lhv is rhv if rhv != 'null' else lhv is None
        if bop.op == 'is not':
            return lhv is not rhv if rhv != 'null' else lhv is not None
        if bop.op == 'in':
            return lhv in rhv
        if bop.op == 'not in':
            return lhv not in rhv
        if bop.op == 'contains':
            return rhv in lhv
        if bop.op == 'startswith':
            return str(lhv).startswith(rhv)
        if bop.op == 'endswith':
            return str(lhv).endswith(rhv)
        if bop.op == 'and':
            return lhv and rhv
        if bop.op == 'or':
            return lhv or rhv

        raise ffd.LogicError(f"Don't know how to handle op: {bop.op}")

    @staticmethod
    def _remove_function_calls(attr: str):
        return list(map(lambda rr: rr[2], regex.findall(r'((\w)\((?R)\))|(\w+)', attr))).pop()

    @staticmethod
    def _parse_attribute_string(attr: AttributeString, data: dict):
        value = data[str(attr)]

        if attr.has_modifiers():
            for modifier in attr.get_modifiers():
                if modifier == 'LOWER':
                    value = value.lower()
                elif modifier == 'UPPER':
                    value = value.upper()

        return value

    def __and__(self, other):
        return BinaryOp(self, 'and', other)

    def __or__(self, other):
        return BinaryOp(self, 'or', other)

    def prune(self, fields: list):
        data = self._prune(fields, self.to_dict())
        if data is INVALID:
            return None
        if data['l'] is INVALID:
            return BinaryOp.from_dict(data['r'])
        if data['r'] is INVALID:
            return BinaryOp.from_dict(data['l'])

        return BinaryOp.from_dict(data)

    def _prune(self, fields: list, data: dict):
        if isinstance(data['l'], dict):
            data['l'] = self._prune(fields, data['l'])
        elif isinstance(data['l'], str) and len(data['l']) > 2:
            prop = data['l']
            if '(' in prop:
                prop = self._remove_function_calls(prop)
            if prop.startswith('a:') and prop[2:] not in fields:
                return INVALID

        if isinstance(data['r'], dict):
            data['r'] = self._prune(fields, data['r'])
        elif isinstance(data['r'], str) and len(data['r']) > 2:
            prop = data['r']
            if '(' in prop:
                prop = self._remove_function_calls(prop)
            if prop.startswith('a:') and prop[2:] not in fields:
                return INVALID

        if data['l'] is INVALID and data['r'] is not INVALID:
            return data['r']
        if data['l'] is not INVALID and data['r'] is INVALID:
            return data['l']
        if data['l'] is INVALID and data['r'] is INVALID:
            return INVALID

        return data

    def to_sql(self, prefix: str = None):
        sql, params, counter = self._to_sql(prefix=prefix)
        return sql, params

    def _to_sql(self, counter: int = None, params: dict = None, prefix: str = None):
        counter = counter or 1
        params = params or {}
        rhv = None
        lhv, params, counter = self._process_op(self.lhv, params, counter, prefix=prefix)
        if self.op == 'is':
            if self.rhv == 'null' or self.rhv is None:
                rhv = 'null'
            elif self.rhv is False or self.rhv is True:
                rhv = str(self.rhv).lower()
        else:
            rhv, params, counter = self._process_op(self.rhv, params, counter, prefix=prefix)

        ret_op = self.op
        if ret_op == 'startswith':
            ret_op = 'like'
            rhv = f"CONCAT({rhv}, '%')"
        elif ret_op == 'endswith':
            ret_op = 'like'
            rhv = f"CONCAT('%', {rhv})"
        elif 'is not' in ret_op and rhv == 'null':
            rhv = 'null'

        return f'({lhv} {ret_op.replace("==", "=")} {rhv})', params, counter

    @staticmethod
    def _process_op(v, params: dict, counter: int, prefix: str = None):
        if isinstance(v, BinaryOp):
            v, p, counter = v._to_sql(counter, params, prefix=prefix)
            params.update(p)
        elif isinstance(v, (list, tuple)):
            placeholders = []
            for i in v:
                var = f'var{counter}'
                counter += 1
                placeholders.append(f':{var}')
                params[var] = i
            v = f'({",".join(placeholders)})'
        elif not isinstance(v, (Attr, AttributeString)):
            var = f'var{counter}'
            counter += 1
            params[var] = v
            v = f':{var}'
        elif isinstance(v, (Attr, AttributeString)) and prefix is not None:
            if isinstance(v, Attr):
                if not v.attr._value.startswith(f'{prefix}."'):
                    v.attr._value = f'{prefix}."{v.attr._value}"'
            else:
                if not v._value.startswith(f'{prefix}."'):
                    v._value = f'{prefix}."{v._value}"'

        return v, params, counter

    def __repr__(self):
        lhv = repr(self.lhv)
        rhv = repr(self.rhv)
        if lhv == '1' and rhv == '1' and self.op == '==':
            return T
        if lhv == T and rhv != T:
            return rhv
        if rhv == T and lhv != T:
            return lhv

        return f'({self.lhv} {self.op} {self.rhv})'.replace('==', '=')

    def __eq__(self, other):
        return isinstance(other, BinaryOp) and self.to_dict() == other.to_dict()



#!/usr/bin/env python

#  Copyright (c) 2019 JD Williams
#
#  This file is part of Firefly, a Python SOA framework built by JD Williams. Firefly is free software; you can
#  redistribute it and/or modify it under the terms of the GNU General Public License as published by the
#  Free Software Foundation; either version 3 of the License, or (at your option) any later version.
#
#  Firefly is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
#  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
#  Public License for more details. You should have received a copy of the GNU Lesser General Public
#  License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#  You should have received a copy of the GNU General Public License along with Firefly. If not, see
#  <http://www.gnu.org/licenses/>.

import os
import sys

# Make sure we're using the version of pylib in this repo, not one installed
# elsewhere on the system.
sys.path.insert(0, os.path.join(os.path.dirname(sys.argv[0]), 'pylib'))
import gyp

if __name__ == '__main__':
  sys.exit(gyp.script_main())



#  Copyright (c) 2019 JD Williams
#
#  This file is part of Firefly, a Python SOA framework built by JD Williams. Firefly is free software; you can
#  redistribute it and/or modify it under the terms of the GNU General Public License as published by the
#  Free Software Foundation; either version 3 of the License, or (at your option) any later version.
#
#  Firefly is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
#  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
#  Public License for more details. You should have received a copy of the GNU Lesser General Public
#  License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#  You should have received a copy of the GNU General Public License along with Firefly. If not, see
#  <http://www.gnu.org/licenses/>.

from __future__ import annotations

import os

import firefly.domain as ffd


@ffd.cli('firefly deploy', alias={'requirements_file': 'r'})
@ffd.command_handler('firefly.Deploy')
class Deploy(ffd.ApplicationService):
    _config: ffd.Configuration = None
    _agent_factory: ffd.AgentFactory = None

    def __call__(self, env: str = 'local', requirements_file: str = None, **kwargs):
        provider = self._config.all.get('provider', 'default')
        if env == 'local':
            provider = 'default'
        deployment = ffd.Deployment(
            environment=env,
            provider=provider,
            project=self._config.all.get('project'),
            requirements_file=requirements_file
        )
        self.dispatch(ffd.DeploymentCreated(deployment=deployment))
        self.dispatch(ffd.DeploymentInitialized(deployment=deployment))

        agent = self._agent_factory(provider)
        self.dispatch(ffd.DeploymentStarting(deployment=deployment))
        agent(deployment, **kwargs)
        self.dispatch(ffd.DeploymentComplete(deployment=deployment))



#  Copyright (c) 2019 JD Williams
#
#  This file is part of Firefly, a Python SOA framework built by JD Williams. Firefly is free software; you can
#  redistribute it and/or modify it under the terms of the GNU General Public License as published by the
#  Free Software Foundation; either version 3 of the License, or (at your option) any later version.
#
#  Firefly is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
#  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
#  Public License for more details. You should have received a copy of the GNU Lesser General Public
#  License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#  You should have received a copy of the GNU General Public License along with Firefly. If not, see
#  <http://www.gnu.org/licenses/>.

from __future__ import annotations

import inspect
from typing import Union, List

import firefly as ff
import firefly.domain.error as error


class On:
    def __call__(self, event: Union[str, type, List[Union[str, type]]]):
        def on_wrapper(cls):
            if isinstance(event, (list, tuple)):
                for e in event:
                    self._apply_event(cls, e)
            else:
                self._apply_event(cls, event)
            return cls

        return on_wrapper

    @staticmethod
    def _apply_event(cls, event: Union[str, type]):
        try:
            cls.add_event(event)
        except AttributeError:
            if inspect.isfunction(cls):
                ff.add_event(cls, event)
            else:
                raise error.FrameworkError('@on used on invalid target')


on = On()



#  Copyright (c) 2019 JD Williams
#
#  This file is part of Firefly, a Python SOA framework built by JD Williams. Firefly is free software; you can
#  redistribute it and/or modify it under the terms of the GNU General Public License as published by the
#  Free Software Foundation; either version 3 of the License, or (at your option) any later version.
#
#  Firefly is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
#  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
#  Public License for more details. You should have received a copy of the GNU Lesser General Public
#  License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#  You should have received a copy of the GNU General Public License along with Firefly. If not, see
#  <http://www.gnu.org/licenses/>.

from __future__ import annotations

import hashlib
from abc import ABC, abstractmethod
from typing import List, TypeVar, Generic, Union, Callable, Optional, Tuple

import firefly.domain as ffd

from ..service.logging.logger import LoggerAware
from ..value_object import GenericBase

T = TypeVar('T')


class Repository(Generic[T], GenericBase, LoggerAware, ABC):
    _serializer: ffd.Serializer = None

    def __init__(self):
        self._entity_hashes = {}
        self._entities = []
        self._deletions = []
        self._parent = None

    @abstractmethod
    def append(self, entity: Union[T, List[T], Tuple[T]], **kwargs):
        pass

    @abstractmethod
    def remove(self, x: Union[T, List[T], Tuple[T], Callable, ffd.BinaryOp], **kwargs):
        pass

    @abstractmethod
    def find(self, x: Union[str, Callable, ffd.BinaryOp], **kwargs) -> Optional[T]:
        pass

    @abstractmethod
    def filter(self, x: Union[Callable, ffd.BinaryOp], **kwargs) -> Repository:
        pass

    @abstractmethod
    def commit(self, **kwargs):
        pass

    @abstractmethod
    def sort(self, cb: Optional[Union[Callable, Tuple[Union[str, Tuple[str, bool]]]]], **kwargs):
        pass

    @abstractmethod
    def clear(self):
        pass

    @abstractmethod
    def destroy(self):
        pass

    def reset(self):
        self._deletions = []
        self._entities = []
        self._entity_hashes = {}

    def touch(self, entity: ffd.Entity):
        if entity.id_value() in self._entity_hashes:
            self._entity_hashes[entity.id_value()] = ''

    @abstractmethod
    def __iter__(self):
        pass

    @abstractmethod
    def __len__(self):
        pass

    @abstractmethod
    def __getitem__(self, item):
        pass

    def _get_search_criteria(self, cb: Union[Callable, ffd.BinaryOp]) -> ffd.BinaryOp:
        if isinstance(cb, ffd.BinaryOp):
            return cb
        return cb(ffd.EntityAttributeSpy(self._type()))

    def _get_hash(self, entity: ffd.Entity):
        return hashlib.md5(self._serializer.serialize(entity.to_dict(force_all=True)).encode('utf-8')).hexdigest()

    def register_entity(self, entity: ffd.Entity):
        self._entity_hashes[entity.id_value()] = self._get_hash(entity)
        self._entities.append(entity)
        if self._parent is not None:
            self._parent.register_entity(entity)

    def _has_changed(self, entity: ffd.Entity):
        if entity.id_value() not in self._entity_hashes:
            return False
        return self._get_hash(entity) != self._entity_hashes[entity.id_value()]

    def _new_entities(self):
        return [e for e in self._entities if e.id_value() not in self._entity_hashes]

    def _changed_entities(self):
        return [e for e in self._entities if self._has_changed(e)]



#  Copyright (c) 2019 JD Williams
#
#  This file is part of Firefly, a Python SOA framework built by JD Williams. Firefly is free software; you can
#  redistribute it and/or modify it under the terms of the GNU General Public License as published by the
#  Free Software Foundation; either version 3 of the License, or (at your option) any later version.
#
#  Firefly is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
#  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
#  Public License for more details. You should have received a copy of the GNU Lesser General Public
#  License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#  You should have received a copy of the GNU General Public License along with Firefly. If not, see
#  <http://www.gnu.org/licenses/>.

from __future__ import annotations

from dataclasses import fields
from typing import List, Any

import firefly.domain.entity as e
import inflection

from .aggregate_root import AggregateRoot
from .entity import id_, required, hidden, IsOneOf, IsMultipleOf, IsLessThanOrEqualTo, IsLessThan, \
    IsGreaterThanOrEqualTo, IsGreaterThan, HasMaxLength, HasMinLength, MatchesPattern, Entity, list_
from .json_schema import JsonSchema
from ..meta.entity_meta import EntityMeta


class MetaAggregate(AggregateRoot):
    id: str = id_()
    schema: JsonSchema = required()
    data: dict = required()
    _obj: AggregateRoot = hidden()

    def __getattr__(self, item):
        try:
            return object.__getattribute__(self, item)
        except AttributeError:
            try:
                return object.__getattribute__(super(), item)
            except AttributeError:
                return getattr(self._obj, item)

    def __setattr__(self, key, value):
        try:
            object.__setattr__(self, key, value)
        except AttributeError:
            try:
                object.__setattr__(super(), key, value)
            except AttributeError:
                setattr(self._obj, key, value)

    def __post_init__(self):
        if self._obj is None:
            self._obj = self._build_entity(self.schema.schema).from_dict(self.data)

    @classmethod
    def _build_entity(cls, schema: dict, name: str = None):
        class_name = name or inflection.camelize(schema['title'].replace(' ', ''))
        props, annotations_ = cls._process_json_schema_properties(schema)
        props['__annotations__'] = annotations_

        return EntityMeta.__new__(EntityMeta, class_name, (Entity,), props, fields_=props, annotations_=annotations_)

    @classmethod
    def _process_json_schema_properties(cls, schema: dict, props: dict = None, annotations_: dict = None):
        props = props or {}
        annotations_ = annotations_ or {}

        for k, v in schema['properties'].items():
            key = inflection.underscore(k)
            prop, annotation = cls._process_property(
                v, 'required' in schema and k in schema['required'], key
            )
            props[key] = prop
            annotations_[key] = annotation

        return props, annotations_

    @classmethod
    def _process_property(cls, schema: dict, required: bool, key: str):
        t = e.required if required else e.optional
        params = {}

        if 'default' in schema:
            params['default'] = schema['default']

        validators = cls._get_validators(schema)
        if len(validators) > 0:
            params['validators'] = validators

        if 'const' in schema:
            return schema['const'], Any

        if 'type' not in schema:
            return t(**params), Any

        if schema['type'] == 'integer':
            return t(**params), int

        if schema['type'] == 'number':
            return t(**params), float

        if schema['type'] == 'null':
            return t(**params), None

        if schema['type'] == 'boolean':
            return t(**params), bool

        if schema['type'] == 'string':
            return t(**params), str

        if schema['type'] == 'array':
            f, type_ = cls._process_property(schema['items'], False, inflection.singularize(key))
            return list_(**params), List[type_]

        if schema['type'] == 'object':
            cls = cls._build_entity(schema, inflection.camelize(key))
            return t(**params), cls

        raise NotImplementedError(f"Don't know how to parse property: {schema}")

    @classmethod
    def _get_validators(cls, schema: dict):
        ret = []

        if 'enum' in schema:
            ret.append(IsOneOf(schema['enum']))

        if 'multipleOf' in schema:
            ret.append(IsMultipleOf(schema['multipleOf']))

        if 'maximum' in schema:
            ret.append(IsLessThanOrEqualTo(schema['maximum']))

        if 'exclusiveMaximum' in schema:
            ret.append(IsLessThan(schema['exclusiveMaximum']))

        if 'minimum' in schema:
            ret.append(IsGreaterThanOrEqualTo(schema['minimum']))

        if 'exclusiveMinimum' in schema:
            ret.append(IsGreaterThan(schema['exclusiveMinimum']))

        if 'maxLength' in schema:
            ret.append(HasMaxLength(schema['maxLength']))

        if 'minLength' in schema:
            ret.append(HasMinLength(schema['minLength']))

        if 'pattern' in schema:
            ret.append(MatchesPattern(schema['pattern']))

        # TODO Array and Object validators
        # https://json-schema.org/draft/2019-09/json-schema-validation.html#rfc.section.6

        return ret



#  Copyright (c) 2019 JD Williams
#
#  This file is part of Firefly, a Python SOA framework built by JD Williams. Firefly is free software; you can
#  redistribute it and/or modify it under the terms of the GNU General Public License as published by the
#  Free Software Foundation; either version 3 of the License, or (at your option) any later version.
#
#  Firefly is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
#  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
#  Public License for more details. You should have received a copy of the GNU Lesser General Public
#  License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#  You should have received a copy of the GNU General Public License along with Firefly. If not, see
#  <http://www.gnu.org/licenses/>.

from __future__ import annotations

import inspect
from pprint import pprint
from typing import Callable, Dict, Type, Union, List

import firefly.domain as ffd
from firefly.domain.entity.messaging.event import Event
from firefly.domain.service.logging.logger import LoggerAware
from firefly.domain.service.messaging.middleware import Middleware


class EventResolvingMiddleware(Middleware, LoggerAware):
    _context_map: ffd.ContextMap = None
    _context: str = None
    _ff_environment: str = None

    def __init__(self, event_listeners: Dict[Union[Type[Event], str], List[ffd.ApplicationService]] = None):
        self._event_listeners = {}
        self._initialized = False
        if event_listeners is not None:
            for event, listeners in event_listeners.items():
                self._event_listeners[event.get_fqn() if not isinstance(event, str) else event] = listeners

    def _initialize(self):
        for event, listeners in self._event_listeners.items():
            built = []
            for listener in listeners:
                if inspect.isclass(listener):
                    built.append(self._context_map.get_context(listener.get_class_context()).container.build(listener))
                else:
                    built.append(listener)
            self._event_listeners[event] = built
        self._initialized = True

    def __call__(self, message: ffd.Message, next_: Callable) -> ffd.Message:
        if not self._initialized:
            self._initialize()

        self.debug('Message context: %s', message.get_context())
        self.debug('This context: %s', self._context)
        self.debug('Environment: %s', self._ff_environment)
        if message.get_context() != 'firefly' and \
                message.get_context() == self._context and \
                not message.headers.get('external', False) and \
                self._ff_environment != 'test':
            self.debug('EventResolvingMiddleware - event originated from this context. Dispatching.')
            self._publish_message(message)
            return next_(message)

        args = message.to_dict(recursive=False)
        args['_message'] = message

        if str(message) in self._event_listeners:
            services = self._event_listeners[str(message)]
            for service in services:
                try:
                    parsed_args = ffd.build_argument_list(args, service)
                    self.debug('Calling service %s with arguments: %s', service.__class__.__name__, parsed_args)
                    service(**parsed_args)
                except TypeError as e:
                    self.exception(e)
                    raise ffd.FrameworkError(f'Error calling {service.__class__.__name__}:\n\n{str(e)}')
        else:
            self.info('No event listener found for message %s', message)

        return next_(message)

    def _publish_message(self, message: ffd.Message):
        self._context_map.get_context(self._context).container.message_transport.dispatch(message)

    def add_event_listener(self, handler: Union[ffd.ApplicationService, Type[ffd.ApplicationService]],
                           event: Union[Type[Event], str]):
        if inspect.isclass(handler):
            handler = self._context_map.get_context(handler.get_class_context()).container.build(handler)
        key = event.get_fqn() if not isinstance(event, str) else event
        if key not in self._event_listeners:
            self._event_listeners[key] = []
        if handler.__class__.__name__ not in list(map(lambda h: h.__class__.__name__, self._event_listeners[key])):

            self._event_listeners[key].append(handler)



#  Copyright (c) 2019 JD Williams
#
#  This file is part of Firefly, a Python SOA framework built by JD Williams. Firefly is free software; you can
#  redistribute it and/or modify it under the terms of the GNU General Public License as published by the
#  Free Software Foundation; either version 3 of the License, or (at your option) any later version.
#
#  Firefly is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
#  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
#  Public License for more details. You should have received a copy of the GNU Lesser General Public
#  License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#  You should have received a copy of the GNU General Public License along with Firefly. If not, see
#  <http://www.gnu.org/licenses/>.

from __future__ import annotations

from typing import Callable

import firefly.domain as ffd
import firefly_di as di

from .authorize_scope import AuthorizeScope


class AuthorizingMiddleware(ffd.Middleware, ffd.ChainOfResponsibility):
    _container: di.Container = None

    def __init__(self):
        super().__init__()
        self._handlers.append(self._container.build(AuthorizeScope))

    def __call__(self, message: ffd.Message, next_: Callable) -> ffd.Message:
        if self.handle(message) is not True:
            raise ffd.UnauthorizedError()
        return next_(message)



#  Copyright (c) 2019 JD Williams
#
#  This file is part of Firefly, a Python SOA framework built by JD Williams. Firefly is free software; you can
#  redistribute it and/or modify it under the terms of the GNU General Public License as published by the
#  Free Software Foundation; either version 3 of the License, or (at your option) any later version.
#
#  Firefly is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
#  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
#  Public License for more details. You should have received a copy of the GNU Lesser General Public
#  License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#  You should have received a copy of the GNU General Public License along with Firefly. If not, see
#  <http://www.gnu.org/licenses/>.

from __future__ import annotations

from abc import abstractmethod, ABC
from dataclasses import fields
from pprint import pprint
from typing import Type, Tuple, Union, List, get_type_hints, Dict

import firefly.domain as ffd
from firefly.infrastructure.repository.rdb_repository import DEFAULT_LIMIT

from ..rdb_storage_interface import RdbStorageInterface


# noinspection PyDataclass
class LegacyStorageInterface(RdbStorageInterface, ffd.LoggerAware, ABC):
    def _generate_select(self, entity_type: Type[ffd.Entity], criteria: ffd.BinaryOp = None, limit: int = None,
                         offset: int = None, sort: Tuple[Union[str, Tuple[str, bool]]] = None, count: bool = False):
        pruned_criteria = None
        indexes = [f.name for f in fields(entity_type)
                   if f.metadata.get('index') is True or f.metadata.get('id') is True]
        if criteria is not None:
            pruned_criteria = criteria.prune(indexes)
            data = {
                'columns': self._select_list(entity_type),
                'criteria': pruned_criteria,
                'count': count and (pruned_criteria == criteria),
            }
        else:
            data = {
                'columns': self._select_list(entity_type),
                'count': count,
            }

        sorted_in_db = False
        if sort is not None:
            sort_fields = []
            for s in sort:
                if str(s[0]) in indexes:
                    sort_fields.append(s)
            data['sort'] = sort_fields
            sorted_in_db = len(sort_fields) == len(sort)

        if (not sort or sorted_in_db) and (pruned_criteria is None or criteria == pruned_criteria):
            if limit is not None:
                data['limit'] = limit

            if offset is not None:
                data['offset'] = offset

        sql, params = self._generate_query(entity_type, f'{self._sql_prefix}/select.sql', data)

        return sql, params, pruned_criteria

    def _all(self, entity_type: Type[ffd.Entity], criteria: ffd.BinaryOp = None, limit: int = None, offset: int = None,
             sort: Tuple[Union[str, Tuple[str, bool]]] = None, raw: bool = False, count: bool = False):
        sql, params, pruned_criteria = self._generate_select(
            entity_type, criteria, limit=limit, offset=offset, sort=sort, count=count
        )

        results = self._execute(sql, params)

        ret = []
        if count and criteria == pruned_criteria:
            return results[0]['c']

        for row in results:
            self.debug('Result row: %s', dict(row))
            ret.append(self._build_entity(entity_type, row, raw=raw))

        if criteria != pruned_criteria:
            if limit is not None and offset is not None and sort is not None:
                self.warning('Paging being performed with non-indexed columns. This may lead to undesirable behavior.')
            ret = list(filter(lambda ee: criteria.matches(ee), ret))
            if count:
                ret = len(ret)

        indexes = [f.name for f in fields(entity_type)
                   if f.metadata.get('index') is True or f.metadata.get('id') is True]
        sorted_in_db = False

        if sort is not None:
            sort_fields = []
            for s in sort:
                if str(s[0]) in indexes:
                    sort_fields.append(s)
            sorted_in_db = len(sort_fields) == len(sort)

        if sort is not None and not sorted_in_db and isinstance(ret, list):
            def sort_keys(x):
                keys = []
                for ss in sort:
                    if len(ss) == 2 and ss[1]:
                        keys.append(-(getattr(x, str(ss[0]))))
                    else:
                        keys.append(getattr(x, str(ss[0])))
                return keys

            ret.sort(key=sort_keys)

        if (sort is not None and not sorted_in_db) and offset is not None and limit is not None:
            if limit == DEFAULT_LIMIT:
                return ret[offset:]
            else:
                return ret[offset:(offset + limit)]

        return ret

    def _build_entity(self, entity: Type[ffd.Entity], data, raw: bool = False):
        if self._map_all is True:
            types = get_type_hints(entity)
            for k, v in data.items():
                t = types[k]
                if ffd.is_type_hint(t):
                    if ffd.get_origin(t) is List:
                        t = list
                    elif ffd.get_origin(t) is Dict:
                        t = dict
                if isinstance(t, type) and issubclass(t, ffd.ValueObject):
                    t = dict

                if (t is dict or t is list) and isinstance(v, str):
                    data[k] = self._serializer.deserialize(v)

        return super()._build_entity(entity, data, raw)

    @abstractmethod
    def _execute(self, sql: str, params: dict = None):
        pass

    @abstractmethod
    def _ensure_connected(self):
        pass

    @abstractmethod
    def _disconnect(self):
        pass



#  Copyright (c) 2019 JD Williams
#
#  This file is part of Firefly, a Python SOA framework built by JD Williams. Firefly is free software; you can
#  redistribute it and/or modify it under the terms of the GNU General Public License as published by the
#  Free Software Foundation; either version 3 of the License, or (at your option) any later version.
#
#  Firefly is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
#  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
#  Public License for more details. You should have received a copy of the GNU Lesser General Public
#  License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#  You should have received a copy of the GNU General Public License along with Firefly. If not, see
#  <http://www.gnu.org/licenses/>.

from __future__ import annotations

from dataclasses import asdict

import firefly.domain as ffd

from ..core.application_service import ApplicationService


class InvokeOn(ApplicationService):
    _message_factory: ffd.MessageFactory = None

    def __init__(self, command_name: str):
        self._command_name = command_name

    def __call__(self, **kwargs):
        return self.invoke(self._message_factory.command(self._command_name, asdict(kwargs['_message'])))



#  Copyright (c) 2019 JD Williams
#
#  This file is part of Firefly, a Python SOA framework built by JD Williams. Firefly is free software; you can
#  redistribute it and/or modify it under the terms of the GNU General Public License as published by the
#  Free Software Foundation; either version 3 of the License, or (at your option) any later version.
#
#  Firefly is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
#  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
#  Public License for more details. You should have received a copy of the GNU Lesser General Public
#  License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#  You should have received a copy of the GNU General Public License along with Firefly. If not, see
#  <http://www.gnu.org/licenses/>.

from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Type, List, Union, Callable, Tuple

import firefly.domain as ffd

from .abstract_storage_interface import AbstractStorageInterface


# noinspection PyDataclass
class DocumentStorageInterface(AbstractStorageInterface, ABC):
    _serializer: ffd.Serializer = None
    _registry: ffd.Registry = None
    _cache: dict = {}

    def __init__(self, **kwargs):
        self._tables_checked = []

    def disconnect(self):
        self._disconnect()

    def _disconnect(self):
        pass

    @abstractmethod
    def _add(self, entity: Union[ffd.Entity, List[ffd.Entity]]):
        pass

    @abstractmethod
    def _all(self, entity_type: Type[ffd.Entity], criteria: ffd.BinaryOp = None, limit: int = None, offset: int = None,
             sort: Tuple[Union[str, Tuple[str, bool]]] = None, raw: bool = False, count: bool = False):
        pass

    @abstractmethod
    def _find(self, uuid: str, entity_type: Type[ffd.Entity]):
        pass

    @abstractmethod
    def _remove(self, entity: Union[ffd.Entity, List[ffd.Entity], Callable]):
        pass

    @abstractmethod
    def _update(self, entity: ffd.Entity):
        pass

    @abstractmethod
    def _ensure_connected(self):
        pass

    def _check_prerequisites(self, entity: Type[ffd.Entity]):
        self._ensure_connected()

    def _build_entity(self, entity: Type[ffd.Entity], data, raw: bool = False):
        if raw is True:
            return self._serializer.deserialize(data)

        data = self._serializer.deserialize(data)

        for k, v in self._get_relationships(entity).items():
            if v['this_side'] == 'one':
                data[k] = self._registry(v['target']).find(data[k])
            elif v['this_side'] == 'many':
                data[k] = list(self._registry(v['target']).filter(
                    lambda ee: getattr(ee, v['target'].id_name()).is_in(data[k])
                ))

        return entity.from_dict(data)



#  Copyright (c) 2019 JD Williams
#
#  This file is part of Firefly, a Python SOA framework built by JD Williams. Firefly is free software; you can
#  redistribute it and/or modify it under the terms of the GNU General Public License as published by the
#  Free Software Foundation; either version 3 of the License, or (at your option) any later version.
#
#  Firefly is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
#  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
#  Public License for more details. You should have received a copy of the GNU Lesser General Public
#  License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#  You should have received a copy of the GNU General Public License along with Firefly. If not, see
#  <http://www.gnu.org/licenses/>.

from __future__ import annotations

from typing import Union, List

from firefly.domain.entity.entity import dict_

# __pragma__('skip')
import uuid

from dataclasses import asdict, fields
# __pragma__('noskip')
# __pragma__('ecom')
"""?
from firefly.presentation.web.polyfills import uuid, asdict, fields
?"""
# __pragma__('noecom')

import firefly.domain as ffd

from firefly.domain.meta.firefly_type import FireflyType
from firefly.domain.meta.message_meta import MessageMeta


class Message(FireflyType, metaclass=MessageMeta):
    headers: dict = dict_()
    _id: str = None
    _context: str = None

    def __init__(self, *args, **kwargs):
        pass

    def get_parameters(self):
        return ffd.get_arguments(self.__init__)

    @property
    def was_external(self):
        return 'client_id' in self.headers

    def __post_init__(self):
        if self._id is None:
            self._id = str(uuid.uuid4())
        if self._context is None:
            if self.__class__._context is not None:
                self._context = self.__class__._context
            else:
                self._context = self.__module__.split('.')[0]

    # noinspection PyDataclass
    def to_dict(self, recursive: bool = True) -> dict:
        if recursive:
            return asdict(self)

        ret = {}
        for field_ in fields(self):
            ret[field_.name] = getattr(self, field_.name)

        return ret
